# Copyright (c) OpenMMLab. All rights reserved.
from collections import defaultdict
from typing import Dict, Optional, Sequence, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.models.decoders.base_decoder import BaseDecoder
from mmflow.registry import MODELS
from mmflow.utils import (OptMultiConfig, OptSampleList, SampleList,
                          TensorDict, TensorList)
from ..builder import build_components, build_loss
from ..utils import BasicDenseBlock, CorrBlock, unpack_flow_data_samples

IRRForwardOutput = Tuple[Dict[str, TensorList], Dict[str, TensorList],
                         Dict[str, TensorList], Dict[str, TensorList]]


class IRRCorrBlock(BaseModule):
    """The correlation block of IRRPWC Net.

    Args:
        in_channels (int): Number of input channels of convolution layers.
        out_channels (int): Number of output channels of convolution layers.
        corr_cfg (dict): Config for correlation layer.
        scaled (bool): Whether to use scaled correlation by the
            number of elements involved to calculate correlation or not.
        warp_cfg(dict): Config for warp operation.
        act_cfg (dict, optional): Config dict for each activation layer in
            ConvModule.
        init_cfg (dict, list, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 out_channels: int,
                 corr_cfg: dict,
                 scaled: bool,
                 warp_cfg: dict,
                 act_cfg: dict,
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(init_cfg)

        if in_channels == out_channels:
            self.conv_1x1 = nn.Sequential()
        else:
            self.conv_1x1 = ConvModule(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=1,
                act_cfg=act_cfg)

        self.corr = CorrBlock(
            corr_cfg=corr_cfg, act_cfg=act_cfg, scaled=scaled)

        self.warp = build_components(warp_cfg)

    def forward(
        self,
        feat1: Tensor,
        feat2: Tensor,
        flow_f: Optional[Tensor] = None,
        flow_b: Optional[Tensor] = None
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor]:
        """Forward function for IRRCorrBlock.

        Args:
            feat1 (Tensor): The feature from image1.
            feat2 (Tensor): The feature from image2.
            flow_f (Optional[Tensor], optional): The optical flow from image1
                to image2, but for level6 there is not predicted flow.
                Defaults to None.
            flow_b (Optional[Tensor], optional): The optical flow from image2
                to image1, but for level6 there is not predicted flow.
                Defaults to None.

        Returns:
            Sequence[Tensor]: correlation between feature1 and warped feature2,
                feature1 after 1x1 convolution layer to make the number of
                channel equal, correlation between feature2 and warped
                feature1, and feature2 after 1x1 convolution layer.
        """

        if flow_f is None and flow_b is None:
            feat1_warp = feat1
            feat2_warp = feat2
        else:
            assert flow_f is not None and flow_b is not None

            feat1_warp = self.warp(feat1, flow_b)
            feat2_warp = self.warp(feat2, flow_f)

        corr_f = self.corr(feat1, feat2_warp)
        corr_b = self.corr(feat2, feat1_warp)

        feat1 = self.conv_1x1(feat1)
        feat2 = self.conv_1x1(feat2)

        return corr_f, feat1, corr_b, feat2


class IRRFlowDecoder(BasicDenseBlock):
    """The decoder works for estimating flow map in IRRPWC.

    Args:
        in_channels (int): Number of input channels per level.
        feat_channels (Sequence[int]): Output channels of convolution module
            in dense layers. Default: (128, 128, 96, 64, 32).
        conv_cfg (dict): Config of convolution layer in module.
        norm_cfg (dict): Config of norm layer in module.
        act_cfg (dict): Config of activation layer in module.
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: Sequence[int],
                 conv_cfg: dict,
                 norm_cfg: dict,
                 act_cfg: dict,
                 init_cfg: Optional[Union[list, dict]] = None) -> None:
        super().__init__(
            in_channels,
            feat_channels=feat_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)
        self.last_channels = self.in_channels + sum(self.feat_channels)
        self.predict_layer = nn.Conv2d(
            self.last_channels, 2, kernel_size=3, padding=1)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function for IRRFlowDecoder.

        Args:
            x (Tensor): The input feature.

        Returns:
            Tensor: The predicted optical flow.
        """
        feat = self.layers(x)
        return feat, self.predict_layer(feat)


class IRROccDecoder(BasicDenseBlock):
    """The decoder works for estimating occlusion map in IRRPWC.

    Args:
        in_channels (int): Number of input channels per level.
        feat_channels (Sequence[int]): Output channels of convolution module
            in dense layers. Default: (128, 128, 96, 64, 32).
        conv_cfg (dict): Config of convolution layer in module.
        norm_cfg (dict): Config of norm layer in module.
        act_cfg (dict): Config of activation layer in module.
        init_cfg (dict, list, optional): Config for module initialization.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: Sequence[int],
                 conv_cfg: dict,
                 norm_cfg: dict,
                 act_cfg: dict,
                 init_cfg: Optional[Union[list, dict]] = None) -> None:
        super().__init__(
            in_channels,
            feat_channels=feat_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            init_cfg=init_cfg)
        self.last_channels = self.in_channels + sum(self.feat_channels)
        self.predict_layer = nn.Conv2d(
            self.last_channels, 1, kernel_size=3, padding=1)

    def forward(self, x: Tensor) -> Tensor:
        """Forward function for IRROccDecoder.

        Args:
            x (Tensor): The input feature.

        Returns:
            Tensor: The predicted occlusion mask.
        """
        feat = self.layers(x)
        return feat, self.predict_layer(feat)


@MODELS.register_module()
class IRRPWCDecoder(BaseDecoder):
    """The decoder module of IRRPWC.

    Args:
        flow_levels (Sequence[str]): The list of output. For multi-levels
            outputs, the flow_levels indicates which levels will be
            predicted.
        corr_in_channels (dict): Dict of input channels of different
            levels of feature pyramid. In IRR-PWC, for reusing correlation
            block for different levels, it needs the same input channels,
            corr_in_channels works for conv1x1 which makes the input
            channels be equal.
        corr_feat_channels (int): Number of output feature channels for
            conv1x1 in corr_block.
        flow_decoder_in_channels (int): Number of input channels that will
            be feed in the flow decoder.
        occ_decoder_in_channels (int): Number of input channels that will
            be feed in the occlusion decoder.
        corr_cfg (dict): Config for correlation layer.
            Defaults to dict(type='Correlation', max_displacement=4).
        scaled (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not.
            Defaults to True.
        warp_cfg (dict): Config for warp operation. Defaults to
            dict(type='Warp', align_corners=True) that are same to the official
            implementation of IRRPWC.
        densefeat_channels (Sequence[int]): Number of output channels for
            dense layers. Defaults to (128, 128, 96, 64, 32).
        flow_post_processor (dict, optional): Config of flow post process
            module. Default: None
        flow_refine (dict, optional): Config of flow refine module.
            Defaults to None.
        occ_post_processor (dict, optional): Config of occlusion post
            process module. Default: None
        occ_refine (dict, optional): Config of occlusion refine module.
            Defaults to None.
        occ_refined_levels (Sequence[str]): List of levels of occlusion
            output that will be refine by `OccShuffleUpsample` module.
            Defaults to ['level0', 'level1'].
        occ_upsample (dict, optional): Config of `OccShuffleUpsample`
            module. Defaults to None.
        flow_div (float): The divisor works for scaling the ground truth.
            Default: 20.
        upsample_cfg (dict): Config dict of interpolate in PyTorch.
            Default: dict(mode='bilinear', align_corners=True)
        conv_cfg (dict, optional): Config dict of convolution layer in
            module. Default: None.
        norm_cfg (dict, optional): Config dict of norm layer in module.
            Default: None.
        act_cfg (dict, optional): Config dict of activation layer in
            module. Default: dict(type='LeakyReLU', negative_slope=0.1).
        flow_loss: Config of loss function of optical flow. Default: None.
        occ_loss: Config of loss function of occlusion mask. Default: None.
        init_cfg (dict, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 flow_levels: Sequence[str],
                 corr_in_channels: Dict[str, int],
                 corr_feat_channels: int,
                 flow_decoder_in_channels: int,
                 occ_decoder_in_channels: int,
                 corr_cfg: dict = dict(type='Correlation', max_displacement=4),
                 scaled: bool = True,
                 warp_cfg: dict = dict(type='Warp', align_corners=True),
                 densefeat_channels: Sequence[int] = (128, 128, 96, 64, 32),
                 flow_post_processor: dict = None,
                 flow_refine: dict = None,
                 occ_post_processor: dict = None,
                 occ_refine: dict = None,
                 occ_refined_levels: Sequence[str] = ['level0', 'level1'],
                 occ_upsample: dict = None,
                 flow_div: float = 20.,
                 upsample_cfg: dict = dict(
                     mode='bilinear', align_corners=True),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 flow_loss: Optional[dict] = None,
                 occ_loss: Optional[dict] = None,
                 init_cfg: OptMultiConfig = None) -> None:

        super().__init__(init_cfg=init_cfg)

        self.flow_levels = flow_levels
        self.flow_levels.sort()
        self.start_level = self.flow_levels[-1]
        self.end_level = self.flow_levels[0]

        if flow_loss is not None:
            self.flow_loss = build_loss(flow_loss)
        if occ_loss is not None:
            self.occ_loss = build_loss(occ_loss)

        self._make_corr_block(
            corr_in_channels=corr_in_channels,
            corr_feat_channels=corr_feat_channels,
            corr_cfg=corr_cfg,
            act_cfg=act_cfg,
            scaled=scaled,
            warp_cfg=warp_cfg)

        self.flow_decoders = IRRFlowDecoder(
            in_channels=flow_decoder_in_channels,
            feat_channels=densefeat_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.occ_decoders = IRROccDecoder(
            in_channels=occ_decoder_in_channels,
            feat_channels=densefeat_channels,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

        self.upsample_cfg = upsample_cfg

        self.flow_refine = build_components(flow_refine)
        self.flow_post_processor = build_components(flow_post_processor)

        self.occ_refined_levels = occ_refined_levels
        self.occ_post_processor = build_components(occ_post_processor)
        self.occ_refine = build_components(occ_refine)
        self.occ_shuffle_upsample = build_components(occ_upsample)

        self.flow_div = flow_div

    def _make_corr_block(self, corr_in_channels: Dict[str, int],
                         corr_feat_channels: int, corr_cfg: dict,
                         act_cfg: dict, scaled: bool, warp_cfg: dict) -> None:
        """Make correlation block.

        Args:
            corr_in_channels (dict): Dict of input channels of different
                levels of feature pyramid. In IRR-PWC, for reusing correlation
                block for different levels, it needs the same input channels,
                corr_in_channels works for conv1x1 which makes the input
                channels be equal.
            corr_feat_channels (int): Number of output feature channels for
                conv1x1 in corr_block.
            corr_cfg (dict): Config for correlation layer.
            act_cfg (dict, optional): Config dict of activation layer in
                module.
            scaled (bool): Whether to use scaled correlation by the number of
                elements involved to calculate correlation or not.
                Defaults to True.
            warp_cfg (dict): Config for warp operation.
        """

        layers = []
        for level, channels in corr_in_channels.items():
            layers.append([
                level,
                IRRCorrBlock(
                    in_channels=channels,
                    out_channels=corr_feat_channels,
                    corr_cfg=corr_cfg,
                    scaled=scaled,
                    warp_cfg=warp_cfg,
                    act_cfg=act_cfg,
                )
            ])
        self.corr_block = nn.ModuleDict(layers)

    def forward(self, feat1: TensorDict,
                feat2: TensorDict) -> IRRForwardOutput:
        """Forward function for IRR-PWC decoder.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
        Returns:
            IRRForwardOutput : The predicted multi-level optical flow and the
                predicted multi-level occlusion mask.
        """

        img1 = feat1['level0']
        img2 = feat2['level0']

        B, _, H_img, W_img = img1.shape

        # pred_flow_fw and pred_flow_bw are the output of optical flow,
        # includes keys:
        #   - levelx: [Tensor]
        #
        # pred_occ_fw and pred_occ_bw are the output of occlusion,
        # includes keys:
        #   - levelx: [Tensor]

        pred_flow_fw = defaultdict(torch.Tensor)
        pred_flow_bw = defaultdict(torch.Tensor)
        pred_occ_fw = defaultdict(torch.Tensor)
        pred_occ_bw = defaultdict(torch.Tensor)

        for level in self.flow_levels[::-1]:

            _feat1, _feat2 = feat1.get(level), feat2.get(level)

            B, _, h, w = _feat1.shape

            if level == self.start_level:

                flow_fw = torch.zeros((B, 2, h, w),
                                      requires_grad=False).to(_feat1)
                flow_bw = torch.zeros((B, 2, h, w),
                                      requires_grad=False).to(_feat1)
                occ_fw = torch.zeros((B, 1, h, w),
                                     requires_grad=False).to(_feat1)
                occ_bw = torch.zeros((B, 1, h, w),
                                     requires_grad=False).to(_feat1)

            else:

                flow_fw = self._scale_flow(flow_fw, h, w)
                flow_bw = self._scale_flow(flow_bw, h, w)
                if level not in self.occ_refined_levels:
                    occ_fw = self._scale_img(occ_fw, h, w)
                    occ_bw = self._scale_img(occ_bw, h, w)

            if level in self.occ_refined_levels:

                pred_flow_fw[level] = [
                    self._scale_flow_as_gt(flow_fw, H_img=H_img, W_img=W_img)
                ]
                pred_flow_bw[level] = [
                    self._scale_flow_as_gt(flow_bw, H_img=H_img, W_img=W_img)
                ]

                occ_fw = F.interpolate(occ_fw, scale_factor=2, mode='nearest')
                occ_bw = F.interpolate(occ_bw, scale_factor=2, mode='nearest')

                occ_fw = self.occ_shuffle_upsample(
                    occ_fw,
                    _feat1,
                    _feat2,
                    flow_fw,
                    flow_bw,
                    self.flow_div,
                    H_img,
                    W_img,
                )
                occ_bw = self.occ_shuffle_upsample(
                    occ_bw,
                    _feat2,
                    _feat1,
                    flow_bw,
                    flow_fw,
                    self.flow_div,
                    H_img,
                    W_img,
                )
                pred_occ_fw[level] = [occ_fw]
                pred_occ_bw[level] = [occ_bw]

            else:
                corr_f, feat1_1by1, corr_b, feat2_1by1 = self.corr_block[
                    level](_feat1, _feat2, flow_fw, flow_bw)

                # predict flow for each level on the original scale
                feat_flow_f, flow_res_f = self.flow_decoders(
                    torch.cat((corr_f, feat1_1by1, flow_fw), dim=1))
                feat_flow_b, flow_res_b = self.flow_decoders(
                    torch.cat((corr_b, feat2_1by1, flow_bw), dim=1))

                flow_fw = flow_res_f + flow_fw
                flow_bw = flow_res_b + flow_bw

                flow_fw = flow_fw + self.flow_post_processor(
                    torch.cat((feat_flow_f, flow_fw), dim=1))
                flow_bw = flow_bw + self.flow_post_processor(
                    torch.cat((feat_flow_b, flow_bw), dim=1))

                # predict occ
                feat_occ_f, occ_res_f = self.occ_decoders(
                    torch.cat((corr_f, feat1_1by1, occ_fw), dim=1))
                feat_occ_b, occ_res_b = self.occ_decoders(
                    torch.cat((corr_b, feat2_1by1, occ_bw), dim=1))
                occ_fw = occ_res_f + occ_fw
                occ_bw = occ_res_b + occ_bw
                occ_fw = occ_fw + self.occ_post_processor(
                    torch.cat((feat_occ_f, occ_fw), dim=1))
                occ_bw = occ_bw + self.occ_post_processor(
                    torch.cat((feat_occ_b, occ_bw), dim=1))

                # refine flow
                scale_img1 = self._scale_img(img1, h, w)
                scale_img2 = self._scale_img(img2, h, w)

                flow_refined_f = self.flow_refine(scale_img1, scale_img2,
                                                  feat1_1by1, flow_fw.detach())
                flow_refined_b = self.flow_refine(scale_img2, scale_img1,
                                                  feat2_1by1, flow_bw.detach())

                # refine occ
                occ_refined_f = self.occ_refine(feat1_1by1, feat2_1by1,
                                                occ_fw.detach(),
                                                flow_refined_f)
                occ_refined_b = self.occ_refine(feat2_1by1, feat1_1by1,
                                                occ_bw.detach(),
                                                flow_refined_b)

                # scale flow as gt scale and divided by flow_div
                # just rescale flow not reshape map
                pred_flow_fw[level] = [
                    self._scale_flow_as_gt(flow_fw, H_img, W_img),
                    self._scale_flow_as_gt(flow_refined_f, H_img, W_img)
                ]
                pred_flow_bw[level] = [
                    self._scale_flow_as_gt(flow_bw, H_img, W_img),
                    self._scale_flow_as_gt(flow_refined_b, H_img, W_img)
                ]
                pred_occ_fw[level] = [occ_fw, occ_refined_f]
                pred_occ_bw[level] = [occ_bw, occ_refined_b]

                flow_fw = flow_refined_f
                flow_bw = flow_refined_b
                occ_fw = occ_refined_f
                occ_bw = occ_refined_b

        return pred_flow_fw, pred_flow_bw, pred_occ_fw, pred_occ_bw

    def _scale_img(self, img: torch.Tensor, h: int, w: int) -> torch.Tensor:
        """Scale image function.

        Args:
            img (Tensor): the input image.
            h (int): the height of output.
            w (int): the width of output.

        Returns:
            Tensor: The output image.
        """
        return F.interpolate(img, size=(h, w), **self.upsample_cfg)

    def _scale_flow(self, flow, h, w):
        """Scale flow function.

        Args:
            flow (Tensor): the input flow.
            h (int): the height of output.
            w (int): the width of output.

        Returns:
            Tensor: The output optical flow.
        """
        h_org, w_org = flow.shape[2:]
        scale = torch.Tensor([float(w / w_org), float(h / h_org)]).to(flow)
        flow = torch.einsum('b c h w, c -> b c h w', flow, scale)

        return F.interpolate(flow, size=(h, w), **self.upsample_cfg)

    def _scale_flow_as_gt(self, flow: torch.Tensor, H_img: int,
                          W_img: int) -> torch.Tensor:
        """Scale optical flow as ground truth.

        Args:
            flow (Tensor): the input flow.
            H_img (int): the height of input images.
            W_img (int): the width of input images.
        Returns:
            Tensor: The output optical flow.
        """
        h_org, w_org = flow.shape[2:]
        scale = torch.Tensor([float(W_img / w_org),
                              float(H_img / h_org)]).to(flow) / self.flow_div
        return torch.einsum('b c h w, c -> b c h w', flow, scale)

    def loss(
        self,
        feat1: TensorDict,
        feat2: TensorDict,
        data_samples: SampleList,
    ) -> TensorDict:
        """Forward function when model training.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """

        pred_flow_fw, pred_flow_bw, pred_occ_fw, pred_occ_bw = self.forward(
            feat1, feat2)
        return self.loss_by_feat(pred_flow_fw, pred_flow_bw, pred_occ_fw,
                                 pred_occ_bw, data_samples)

    def predict(self,
                feat1: TensorDict,
                feat2: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model testing.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        pred_flow_fw, pred_flow_bw, _, _ = self.forward(feat1, feat2)

        flow_results_fw = pred_flow_fw[self.end_level][-1]
        flow_results_bw = pred_flow_bw[self.end_level][-1]
        return self.predict_by_feat(flow_results_fw, flow_results_bw,
                                    data_samples)

    def predict_by_feat(self,
                        flow_results_fw: Tensor,
                        flow_results_bw: Tensor,
                        data_samples: OptSampleList = None) -> SampleList:
        """Predict list of obj:`FlowDataSample` from flow tensor.

        Args:
            flow_results_fw (Tensor): Predicted forward flow tensor.
            flow_results_bw (Tensor): predicted backward flow tensor.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        if data_samples is None:
            flow_results_fw = flow_results_fw * self.flow_div
            flow_results_bw = flow_results_bw * self.flow_div

            # unravel batch dim,
            flow_results_fw = list(flow_results_fw)
            flow_results_bw = list(flow_results_bw)

            flow_results = [
                dict(flow_fw=f, flow_bw=b)
                for f, b in zip(flow_results_fw, flow_results_bw)
            ]
            return self.postprocess_result(flow_results, data_samples=None)

        H, W = data_samples[0].metainfo['img_shape'][:2]

        flow_results_fw = F.interpolate(
            flow_results_fw, size=(H, W), mode='bilinear', align_corners=False)
        flow_results_fw = flow_results_fw * self.flow_div
        # unravel batch dim
        flow_results_fw = list(flow_results_fw)

        flow_results_bw = F.interpolate(
            flow_results_bw, size=(H, W), mode='bilinear', align_corners=False)
        flow_result_bw = flow_results_bw * self.flow_div
        # unravel batch dim
        flow_result_bw = list(flow_result_bw)

        # collect forward and backward flow
        results = [
            dict(flow_fw=f, flow_bw=b)
            for f, b in zip(flow_results_fw, flow_results_bw)
        ]

        return self.postprocess_result(results, data_samples=data_samples)

    def loss_by_feat(
        self,
        pred_flow_fw: Dict[str, Sequence[torch.Tensor]],
        pred_flow_bw: Dict[str, Sequence[torch.Tensor]],
        pred_occ_fw: Dict[str, Sequence[torch.Tensor]],
        pred_occ_bw: Dict[str, Sequence[torch.Tensor]],
        data_samples: SampleList,
    ) -> TensorDict:
        """Compute optical flow loss and occlusion mask loss.

        Args:
            pred_flow_fw (dict): multi-level predicted forward optical flow.
            pred_flow_bw (dict): multi-level predicted backward optical flow.
            pred_occ_fw (dict): multi-level predicted forward occlusion mask.
            pred_occ_bw (dict): multi-level predicted backward occlusion mask.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """
        batch_gt_flow_fw, batch_gt_flow_bw, batch_gt_occ_fw, \
            batch_gt_occ_bw, batch_gt_valid_fw, batch_gt_valid_bw = \
            unpack_flow_data_samples(data_samples)

        losses = dict()

        def _flow_loss_or_detach(preds, gts, valids):
            if gts is None:
                self._detach_unused_preds(preds)
                return None
            else:
                # only sintel and kitti dataset have valid mask, neither of
                # them don't have backward flow ground truth
                return self.flow_loss(preds, gts, valid=valids)

        def _occ_loss_or_detach(preds, gts):
            if gts is None:
                self._detach_unused_preds(preds)
                return None
            else:
                return self.occ_loss(preds, gts)

        n_flow_loss = 0
        losses['loss_flow'] = 0
        for pred_flow, gt_flow, gt_valid in zip(
            (pred_flow_fw, pred_flow_bw),
            (batch_gt_flow_fw, batch_gt_flow_bw),
            (batch_gt_valid_fw, batch_gt_valid_bw),
        ):

            loss_flow = _flow_loss_or_detach(pred_flow, gt_flow, gt_valid)
            if loss_flow is not None:
                losses['loss_flow'] += loss_flow
                n_flow_loss += 1
        if n_flow_loss > 0:
            losses['loss_flow'] = losses['loss_flow'] / n_flow_loss

        n_occ_loss = 0
        losses['loss_occ'] = 0
        for pred_occ, gt_occ in zip((pred_occ_fw, pred_occ_bw),
                                    (batch_gt_occ_fw, batch_gt_occ_bw)):
            loss_occ = _occ_loss_or_detach(pred_occ, gt_occ)
            if loss_occ is not None:
                losses['loss_occ'] += loss_occ
                n_occ_loss += 1
        if n_occ_loss > 0:
            losses['loss_occ'] = losses['loss_occ'] / n_occ_loss

        return losses

    @staticmethod
    def _detach_unused_preds(preds: TensorDict) -> None:
        """Detach unused predicted output.

        As some datasets do not include ground truth of backward optical flow
        and occlusion, this function detaches the prediction that don't need
        to calculate the loss.

        Args:
            preds (Dict[str, Tensor]): The prediction without corresponding
                target.
        """
        for k in preds.keys():
            for i in range(len(preds[k])):
                preds[k][i] = preds[k][i].detach()



# Copyright (c) OpenMMLab. All rights reserved.
import torch.nn as nn
import torch.utils.checkpoint as cp
from mmcv.cnn import (build_activation_layer, build_conv_layer,
                      build_norm_layer, build_plugin_layer)
from mmengine.model import BaseModule


class BasicBlock(BaseModule):
    """Basic block for ResNet."""

    expansion = 1

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 dcn=None,
                 plugins=None):
        super(BasicBlock, self).__init__()
        assert dcn is None, 'Not implemented yet.'
        assert plugins is None, 'Not implemented yet.'

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            3,
            stride=stride,
            padding=1,
            dilation=dilation,
            bias=True)
        self.add_module(self.norm1_name, norm1)
        self.conv2 = build_conv_layer(
            conv_cfg, planes, planes, 3, padding=1, bias=True)
        self.add_module(self.norm2_name, norm2)

        self.relu = build_activation_layer(act_cfg)
        self.downsample = downsample
        self.stride = stride
        self.dilation = dilation
        self.with_cp = with_cp

    @property
    def norm1(self):
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    def forward(self, x):
        """Forward function."""

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.relu(out)

            out = self.conv2(out)
            out = self.norm2(out)
            out = self.relu(out)

            if self.downsample is not None:
                identity = self.downsample(x)

            out = out + identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        return out


class Bottleneck(BaseModule):
    """Bottleneck block for ResNet.

    If style is "pytorch", the stride-two layer is the 3x3 conv layer, if it is
    "caffe", the stride-two layer is the first 1x1 conv layer.
    """

    expansion = 4

    def __init__(self,
                 inplanes,
                 planes,
                 stride=1,
                 dilation=1,
                 downsample=None,
                 style='pytorch',
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 dcn=None,
                 plugins=None):
        super(Bottleneck, self).__init__()
        assert style in ['pytorch', 'caffe']
        assert dcn is None or isinstance(dcn, dict)
        assert plugins is None or isinstance(plugins, list)
        if plugins is not None:
            allowed_position = ['after_conv1', 'after_conv2', 'after_conv3']
            assert all(p['position'] in allowed_position for p in plugins)

        self.inplanes = inplanes
        self.planes = planes
        self.stride = stride
        self.dilation = dilation
        self.style = style
        self.with_cp = with_cp
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.dcn = dcn
        self.with_dcn = dcn is not None
        self.plugins = plugins
        self.with_plugins = plugins is not None

        if self.with_plugins:
            # collect plugins for conv1/conv2/conv3
            self.after_conv1_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv1'
            ]
            self.after_conv2_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv2'
            ]
            self.after_conv3_plugins = [
                plugin['cfg'] for plugin in plugins
                if plugin['position'] == 'after_conv3'
            ]

        if self.style == 'pytorch':
            self.conv1_stride = 1
            self.conv2_stride = stride
        else:
            self.conv1_stride = stride
            self.conv2_stride = 1

        self.norm1_name, norm1 = build_norm_layer(norm_cfg, planes, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(norm_cfg, planes, postfix=2)
        self.norm3_name, norm3 = build_norm_layer(
            norm_cfg, planes * self.expansion, postfix=3)

        self.conv1 = build_conv_layer(
            conv_cfg,
            inplanes,
            planes,
            kernel_size=1,
            stride=self.conv1_stride,
            bias=True)
        self.add_module(self.norm1_name, norm1)
        fallback_on_stride = False
        if self.with_dcn:
            fallback_on_stride = dcn.pop('fallback_on_stride', False)
        if not self.with_dcn or fallback_on_stride:
            self.conv2 = build_conv_layer(
                conv_cfg,
                planes,
                planes,
                kernel_size=3,
                stride=self.conv2_stride,
                padding=dilation,
                dilation=dilation,
                bias=True)
        else:
            assert self.conv_cfg is None, 'conv_cfg must be None for DCN'
            self.conv2 = build_conv_layer(
                dcn,
                planes,
                planes,
                kernel_size=3,
                stride=self.conv2_stride,
                padding=dilation,
                dilation=dilation,
                bias=False)

        self.add_module(self.norm2_name, norm2)
        self.conv3 = build_conv_layer(
            conv_cfg,
            planes,
            planes * self.expansion,
            kernel_size=1,
            bias=True)
        self.add_module(self.norm3_name, norm3)

        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample

        if self.with_plugins:
            self.after_conv1_plugin_names = self.make_block_plugins(
                planes, self.after_conv1_plugins)
            self.after_conv2_plugin_names = self.make_block_plugins(
                planes, self.after_conv2_plugins)
            self.after_conv3_plugin_names = self.make_block_plugins(
                planes * self.expansion, self.after_conv3_plugins)

    def make_block_plugins(self, in_channels, plugins):
        """make plugins for block.

        Args:
            in_channels (int): Input channels of plugin.
            plugins (list[dict]): List of plugins cfg to build.

        Returns:
            list[str]: List of the names of plugin.
        """
        assert isinstance(plugins, list)
        plugin_names = []
        for plugin in plugins:
            plugin = plugin.copy()
            name, layer = build_plugin_layer(
                plugin,
                in_channels=in_channels,
                postfix=plugin.pop('postfix', ''))
            assert not hasattr(self, name), f'duplicate plugin {name}'
            self.add_module(name, layer)
            plugin_names.append(name)
        return plugin_names

    def forward_plugin(self, x, plugin_names):
        """Forward function for plugins."""
        out = x
        for name in plugin_names:
            out = getattr(self, name)(x)
        return out

    @property
    def norm1(self):
        """nn.Module: normalization layer after the first convolution layer"""
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        """nn.Module: normalization layer after the second convolution layer"""
        return getattr(self, self.norm2_name)

    @property
    def norm3(self):
        """nn.Module: normalization layer after the third convolution layer"""
        return getattr(self, self.norm3_name)

    def forward(self, x):
        """Forward function."""

        def _inner_forward(x):
            identity = x

            out = self.conv1(x)
            out = self.norm1(out)
            out = self.relu(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv1_plugin_names)

            out = self.conv2(out)
            out = self.norm2(out)
            out = self.relu(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv2_plugin_names)

            out = self.conv3(out)
            out = self.norm3(out)

            if self.with_plugins:
                out = self.forward_plugin(out, self.after_conv3_plugin_names)

            if self.downsample is not None:
                identity = self.downsample(x)

            out += identity

            return out

        if self.with_cp and x.requires_grad:
            out = cp.checkpoint(_inner_forward, x)
        else:
            out = _inner_forward(x)

        out = self.relu(out)

        return out


class ResLayer(nn.Sequential):
    """ResLayer to build ResNet style backbone.

    Args:
        block (nn.Module): block used to build ResLayer.
        inplanes (int): inplanes of block.
        planes (int): planes of block.
        num_blocks (int): number of blocks.
        stride (int): stride of the first block. Default: 1
        avg_down (bool): Use AvgPool instead of stride conv when
            downsampling in the bottleneck. Default: False
        conv_cfg (dict): dictionary to construct and config conv layer.
            Default: None
        norm_cfg (dict): dictionary to construct and config norm layer.
            Default: dict(type='BN')
        multi_grid (int | None): Multi grid dilation rates of last
            stage. Default: None
        contract_dilation (bool): Whether contract first dilation of each layer
            Default: False
    """

    def __init__(self,
                 block,
                 inplanes,
                 planes,
                 num_blocks,
                 stride=1,
                 dilation=1,
                 avg_down=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 multi_grid=None,
                 contract_dilation=False,
                 **kwargs):
        self.block = block

        downsample = None
        if stride != 1 or inplanes != planes * block.expansion:
            downsample = []
            conv_stride = stride
            if avg_down:
                conv_stride = 1
                downsample.append(
                    nn.AvgPool2d(
                        kernel_size=stride,
                        stride=stride,
                        ceil_mode=True,
                        count_include_pad=False))
            downsample.extend([
                build_conv_layer(
                    conv_cfg,
                    inplanes,
                    planes * block.expansion,
                    kernel_size=1,
                    stride=conv_stride,
                    bias=True),
                build_norm_layer(norm_cfg, planes * block.expansion)[1]
            ])
            downsample = nn.Sequential(*downsample)

        layers = []
        if multi_grid is None:
            if dilation > 1 and contract_dilation:
                first_dilation = dilation // 2
            else:
                first_dilation = dilation
        else:
            first_dilation = multi_grid[0]
        layers.append(
            block(
                inplanes=inplanes,
                planes=planes,
                stride=stride,
                dilation=first_dilation,
                downsample=downsample,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                **kwargs))
        inplanes = planes * block.expansion
        for i in range(1, num_blocks):
            layers.append(
                block(
                    inplanes=inplanes,
                    planes=planes,
                    stride=1,
                    dilation=dilation if multi_grid is None else multi_grid[i],
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    **kwargs))
        super(ResLayer, self).__init__(*layers)



# Copyright (c) OpenMMLab. All rights reserved.
import math
from typing import Iterator, Optional, Sequence

import torch
import torch.distributed as dist
from mmengine.dataset import ConcatDataset
from torch.utils.data import Dataset
from torch.utils.data import DistributedSampler as _DistributedSampler
from torch.utils.data import Sampler

from mmflow.registry import DATA_SAMPLERS
from mmflow.utils import sync_random_seed


@DATA_SAMPLERS.register_module()
class DistributedSampler(_DistributedSampler):
    """DistributedSampler inheriting from
    `torch.utils.data.DistributedSampler`.

    This distributed sampler is compatible Pytorch==1.5, as there is no
    `seed` argument in Pytorch==1.5.

    Args:
        datasets (Dataset): the dataset will be loaded.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, world_size is retrieved from the
            current distributed group.
        rank (int, optional):  Rank of the current process within num_replicas.
            By default, rank is retrieved from the current distributed group.
        shuffle (bool): If True (default), sampler will shuffle the indices.
        seed (int): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.
    """

    def __init__(self,
                 dataset: Dataset,
                 num_replicas: Optional[int] = None,
                 rank: Optional[int] = None,
                 shuffle: bool = True,
                 seed=0) -> None:
        super().__init__(
            dataset, num_replicas=num_replicas, rank=rank, shuffle=shuffle)

        # In distributed sampling, different ranks should sample
        # non-overlapped data in the dataset. Therefore, this function
        # is used to make sure that each rank shuffles the data indices
        # in the same order based on the same seed. Then different
        # ranks could use different indices to select non-overlapped
        # data from the same data list.
        self.seed = sync_random_seed(seed)

    def __iter__(self) -> Iterator:
        """
         Yields:
            Iterator: iterator of indices for rank.
        """
        # deterministically shuffle based on epoch
        if self.shuffle:
            g = torch.Generator()
            # When :attr:`shuffle=True`, this ensures all replicas
            # use a different random ordering for each epoch.
            # Otherwise, the next iteration of this sampler will
            # yield the same ordering.
            g.manual_seed(self.epoch + self.seed)
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = torch.arange(len(self.dataset)).tolist()

        # add extra samples to make it evenly divisible
        indices += indices[:(self.total_size - len(indices))]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        return iter(indices)


@DATA_SAMPLERS.register_module()
class MixedBatchDistributedSampler(Sampler):
    """Distributed Sampler for mixed data batch.

    Args:
        dataset (ConcatDataset): Dataset that will be loaded.
        sample_ratio (list): List of the ratio of each dataset in a batch, e.g.
            datasets=[DatasetA, DatasetB], sample_ratio=[0.25, 0.75],
            sample_per_gpu=1, gpus=8, it means 2 gpus load DatasetA, and 6 gpus
            load DatasetB. The length of datasets must be equal to length of
            sample_ratio.
        num_replicas (int, optional): Number of processes participating in
            distributed training. By default, world_size is retrieved from the
            current distributed group.
        rank (int, optional):  Rank of the current process within num_replicas.
            By default, rank is retrieved from the current distributed group.
        shuffle (bool): If True (default), sampler will shuffle the indices.
        seed (int): random seed used to shuffle the sampler if
            :attr:`shuffle=True`. This number should be identical across all
            processes in the distributed group. Default: ``0``.
    """

    def __init__(self,
                 dataset: ConcatDataset,
                 sample_ratio: Sequence[float],
                 num_replicas: Optional[int] = None,
                 rank: Optional[int] = None,
                 shuffle: bool = True,
                 seed: int = 0) -> None:

        # base class `Sampler` do nothing in `__init__` function
        # super().__init__()

        datasets = dataset.datasets
        assert len(datasets) == len(sample_ratio)
        assert sum(sample_ratio) == 1.

        if num_replicas is None:
            if not dist.is_available():
                raise RuntimeError(
                    'Requires distributed package to be available')
            num_replicas = dist.get_world_size()

        if rank is None:
            if not dist.is_available():
                raise RuntimeError(
                    'Requires distributed package to be available')
            rank = dist.get_rank()

        if rank >= num_replicas or rank < 0:
            raise ValueError(
                f'Invalid rank {rank}, rank should be in the interval'
                f' [0, {num_replicas - 1}]')

        self.datasets = datasets
        self.num_replicas = num_replicas

        self.datasets_num_replicas = [
            math.ceil(num_replicas * r) for r in sample_ratio
        ]
        self.cumulative_replicas = []
        t = 0
        for nr in self.datasets_num_replicas:
            t += nr
            self.cumulative_replicas.append(t)
        self.datasets_length = [len(d) for d in datasets]
        self.cumulative_datasets_length = []

        t = 0
        for dl in self.datasets_length:
            t += dl
            self.cumulative_datasets_length.append(t)

        # the smallest num_sample
        self.datasets_num_samples = [
            math.ceil(length / ratio) for length, ratio in zip(
                self.datasets_length, self.datasets_num_replicas)
        ]
        self.num_samples = min(self.datasets_num_samples)

        # the dataset that decides the num_samples and total_size
        self.key_dataset = self.datasets_num_samples.index(self.num_samples)
        self.key_dataset_length = self.datasets_length[self.key_dataset]
        self.total_size = [
            self.num_samples * nr for nr in self.datasets_num_replicas
        ]

        self.rank = rank
        self.epoch = 0
        self.shuffle = shuffle
        # In distributed sampling, different ranks should sample
        # non-overlapped data in the dataset. Therefore, this function
        # is used to make sure that each rank shuffles the data indices
        # in the same order based on the same seed. Then different
        # ranks could use different indices to select non-overlapped
        # data from the same data list.
        self.seed = sync_random_seed(seed)

    def __iter__(self) -> Iterator:
        """
        Yields:
            Iterator: iterator of indices for current rank.
        """

        # datasets map different rank
        for dataset_idx, cumulative_replicas_ in enumerate(
                self.cumulative_replicas):
            if self.rank < cumulative_replicas_:
                break

        # deterministically shuffle each datasets based on epoch
        if self.shuffle:
            g = torch.Generator()
            # When :attr:`shuffle=True`, this ensures all replicas
            # use a different random ordering for each epoch.
            # Otherwise, the next iteration of this sampler will
            # yield the same ordering.
            g.manual_seed(self.epoch + self.seed)

            indices = torch.randperm(
                self.datasets_length[dataset_idx], generator=g).tolist()
        else:
            indices = torch.arange(self.datasets_length[dataset_idx]).tolist()

        if self.total_size[dataset_idx] > len(indices):
            # add extra samples for key_dataset to make it evenly divisible
            indices += indices[:(self.total_size[dataset_idx] - len(indices))]
        else:
            indices = indices[:self.total_size[dataset_idx]]

        assert len(indices) == self.total_size[dataset_idx]

        # subsample
        last_cumulative_replicas = 0 \
            if dataset_idx == 0 else self.cumulative_replicas[dataset_idx - 1]

        indices = indices[(
            self.rank - last_cumulative_replicas
        ):self.total_size[dataset_idx]:self.datasets_num_replicas[dataset_idx]]

        assert len(indices) == self.num_samples

        # find the dataset for this rank
        last_cumulative_length = 0 \
            if dataset_idx == 0 else \
            self.cumulative_datasets_length[dataset_idx-1]

        indices = [idx + last_cumulative_length for idx in indices]

        return iter(indices)

    def __len__(self) -> int:
        """Get the combined dataset length."""
        return self.num_samples

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for this sampler. When :attr:`shuffle=True`, this
        ensures all replicas use a different random ordering for each epoch.
        Otherwise, the next iteration of this sampler will yield the same
        ordering.

        Arguments:
            epoch (int): Epoch number.
        """
        self.epoch = epoch



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict

import torch
from torch import Tensor

from ..builder import build_components


def flow_to_coords(flow: Tensor) -> Tensor:
    """Generate shifted coordinate grid based based input flow.

    Args:
        flow (Tensor): Estimated optical flow.

    Returns:
        Tensor: Coordinate that shifted by input flow with shape (B, 2, H, W).
    """
    B, _, H, W = flow.shape
    xx = torch.arange(0, W, device=flow.device, requires_grad=False)
    yy = torch.arange(0, H, device=flow.device, requires_grad=False)
    coords = torch.meshgrid(yy, xx)
    coords = torch.stack(coords[::-1], dim=0).float()
    coords = coords[None].repeat(B, 1, 1, 1) + flow
    return coords


def compute_range_map(flow: Tensor) -> Tensor:
    """Compute range map.

    Args:
        flow (Tensor): The backward flow with shape (N, 2, H, W)

    Return:
        Tensor: The forward-to-backward occlusion mask with shape (N, 1, H, W)
    """

    N, _, H, W = flow.shape

    coords = flow_to_coords(flow)

    # Split coordinates into an integer part and
    # a float offset for interpolation.
    coords_floor = torch.floor(coords)
    coords_offset = coords - coords_floor
    coords_floor = coords_floor.to(torch.int32)

    # Define a batch offset for flattened indexes into all pixels.
    batch_range = torch.arange(N).view(N, 1, 1)
    idx_batch_offset = batch_range.repeat(1, H, W) * H * W

    # Flatten everything.
    coords_floor_flattened = coords_floor.permute(0, 2, 3, 1).reshape(-1, 2)
    coords_offset_flattened = coords_offset.permute(0, 2, 3, 1).reshape(-1, 2)
    idx_batch_offset_flattened = idx_batch_offset.reshape(-1)

    # Initialize results.
    idxs_list = []
    weights_list = []

    # Loop over differences di and dj to the four neighboring pixels.
    for di in range(2):
        for dj in range(2):
            # Compute the neighboring pixel coordinates.
            idxs_j = coords_floor_flattened[..., 0] + dj
            idxs_i = coords_floor_flattened[..., 1] + di
            # Compute the flat index into all pixels.
            idxs = idx_batch_offset_flattened + idxs_i * W + idxs_j

            # Only count valid pixels.
            mask = torch.logical_and(
                torch.logical_and(idxs_j >= 0, idxs_j < W),
                torch.logical_and(idxs_i >= 0, idxs_i < H))
            valid_idxs = idxs[mask]
            valid_offsets = coords_offset_flattened[mask]

            # Compute weights according to bilinear interpolation.
            weights_j = (1. - dj) - (-1)**dj * valid_offsets[:, 0]
            weights_i = (1. - di) - (-1)**di * valid_offsets[:, 1]
            weights = weights_i * weights_j

            # Append indices and weights to the corresponding list.
            idxs_list.append(valid_idxs)
            weights_list.append(weights)
    # Concatenate everything.
    idxs = torch.cat(idxs_list, dim=0)
    weights = torch.cat(weights_list, dim=0)

    # Sum up weights for each pixel and reshape the result.
    count_image = torch.zeros(N * H * W)
    count_image = count_image.index_add_(
        dim=0, index=idxs, source=weights).reshape(N, H, W)
    occ = (count_image >= 1).to(flow)[:, None, ...]
    return occ


def forward_backward_consistency(
        flow_fw: Tensor,
        flow_bw: Tensor,
        warp_cfg: dict = dict(type='Warp', align_corners=True),
) -> Tensor:
    """Occlusion mask from forward-backward consistency.

    Args:
        flow_fw (Tensor): The forward flow with shape (N, 2, H, W)
        flow_bw (Tensor): The backward flow with shape (N, 2, H, W)

    Returns:
        Tensor: The forward-to-backward occlusion mask with shape (N, 1, H, W)
    """

    warp = build_components(warp_cfg)

    warped_flow_bw = warp(flow_bw, flow_fw)

    forward_backward_sq_diff = torch.sum(
        (flow_fw + warped_flow_bw)**2, dim=1, keepdim=True)
    forward_backward_sum_sq = torch.sum(
        flow_fw * 2 + warped_flow_bw**2, dim=1, keepdim=True)

    occ = (forward_backward_sq_diff <
           forward_backward_sum_sq * 0.01 + 0.5).to(flow_fw)
    return occ


def forward_backward_absdiff(flow_fw: Tensor,
                             flow_bw: Tensor,
                             warp_cfg: dict = dict(
                                 type='Warp', align_corners=True),
                             diff: int = 1.5) -> Tensor:
    """Occlusion mask from forward-backward consistency.

    Args:
        flow_fw (Tensor): The forward flow with shape (N, 2, H, W)
        flow_bw (Tensor): The backward flow with shape (N, 2, H, W)

    Returns:
        Tensor: The forward-to-backward occlusion mask with shape (N, 1, H, W)
    """

    warp = build_components(warp_cfg)

    warped_flow_bw = warp(flow_bw, flow_fw)

    forward_backward_sq_diff = torch.sum(
        (flow_fw + warped_flow_bw)**2, dim=1, keepdim=True)

    occ = (forward_backward_sq_diff**0.5 < diff).to(flow_fw)

    return occ


def occlusion_estimation(flow_fw: Tensor,
                         flow_bw: Tensor,
                         mode: str = 'consistency',
                         **kwarg) -> Dict[str, Tensor]:
    """Occlusion estimation.

    Args:
        flow_fw (Tensor): The forward flow with shape (N, 2, H, W)
        flow_bw (Tensor): The backward flow with shape (N, 2, H, W)
        mode (str): The method for occlusion estimation, which can be
            ``'consistency'``, ``'range_map'`` or ``'fb_abs'``.
        warp_cfg (dict, optional): _description_. Defaults to None.

    Returns:
        Dict[str,Tensor]: 1 denote non-occluded and 0 denote occluded
    """
    assert mode in ('consistency', 'range_map', 'fb_abs'), \
        'mode must be \'consistency\', \'range_map\' or \'fb_abs\', ' \
        f'but got {mode}'

    if mode == 'consistency':
        occ_fw = forward_backward_consistency(flow_fw, flow_bw, **kwarg)
        occ_bw = forward_backward_consistency(flow_bw, flow_fw, **kwarg)

    elif mode == 'range_map':
        occ_fw = compute_range_map(flow_bw)
        occ_bw = compute_range_map(flow_fw)

    elif mode == 'fb_abs':
        occ_fw = forward_backward_absdiff(flow_fw, flow_bw, **kwarg)
        occ_bw = forward_backward_absdiff(flow_bw, flow_fw, **kwarg)

    return dict(occ_fw=occ_fw, occ_bw=occ_bw)



# Copyright (c) OpenMMLab. All rights reserved.
from .census_loss import census_loss
from .multilevel_bce import (MultiLevelBCE, binary_cross_entropy,
                             multi_levels_binary_cross_entropy)
from .multilevel_charbonnier_loss import (MultiLevelCharbonnierLoss,
                                          charbonnier_loss)
from .multilevel_epe import MultiLevelEPE, endpoint_error
from .sequence_loss import SequenceLoss, sequence_loss
from .smooth_loss import smooth_1st_loss, smooth_2nd_loss
from .ssim import weighted_ssim

__all__ = [
    'endpoint_error', 'sequence_loss', 'binary_cross_entropy', 'SequenceLoss',
    'MultiLevelBCE', 'MultiLevelEPE', 'MultiLevelCharbonnierLoss',
    'multi_levels_binary_cross_entropy', 'charbonnier_loss', 'weighted_ssim',
    'smooth_1st_loss', 'smooth_2nd_loss', 'census_loss'
]



# Copyright (c) OpenMMLab. All rights reserved.
import torch
import torch.nn as nn
from torch import Tensor

from ..builder import build_components


class LinkOutput:
    """The link output between two estimators in FlowNet2."""

    def __init__(self) -> None:
        self.upsample_flow = None
        self.scaled_flow = None
        self.norm_scaled_flow = None
        self.warped_img2 = None
        self.diff = None
        self.brightness_err = None


class BasicLink(nn.Module):
    """Connect two separate flow estimators.

    BasicLink compute the following 5 values: upsampled flow prediction,
    normalized upsampled flow prediction, warped image, difference between the
    first image and warped image, brightness error.

    Args:
        scale_factor (int): Scale factor of upsampling. Default to 4.
        mode (str): Algorithm used for upsampling: 'nearest' , 'linear' ,
            'bilinear' , 'bicubic' , 'trilinear' , 'area'. Default: 'bilinear'.
        warp_cfg (dict): Config for warp operator. Default to
            dict(type='Warp', padding_mode='border', align_corners=True))
    """

    def __init__(self,
                 scale_factor: int = 4,
                 mode: str = 'bilinear',
                 warp_cfg: dict = dict(
                     type='Warp', padding_mode='border', align_corners=True)):
        super(BasicLink, self).__init__()

        self.warp = build_components(warp_cfg)
        self.upSample = nn.Upsample(scale_factor=scale_factor, mode=mode)

    def __call__(self, img1: Tensor, img2: Tensor, flow: Tensor,
                 flow_div: float) -> LinkOutput:
        """Call function for BasicLink.

        Args:
            img1 (Tensor): The first input image.
            img2 (Tensor): The second input images.
            flow (Tensor): The estimated optical flow from the first image to
                the second image.
            flow_div (float): The divisor for scaling the value of optical
                flow.

        Returns:
            LinkOutput: The output for the next flow estimator.
        """
        upsample_flow = self.upSample(flow)
        scaled_flow = self.upSample(flow) * flow_div
        norm_scaled_flow = torch.norm(scaled_flow, p=2, dim=1, keepdim=True)
        warped_img2 = self.warp(img2, scaled_flow)
        diff = img1 - warped_img2
        bright_err = torch.norm(diff, p=2, dim=1, keepdim=True)

        output = LinkOutput()
        output.upsample_flow = upsample_flow
        output.scaled_flow = scaled_flow
        output.norm_scaled_flow = norm_scaled_flow
        output.warped_img2 = warped_img2
        output.diff = diff
        output.brightness_err = bright_err

        return output



_base_ = [
    '../_base_/models/flownetc.py', '../_base_/datasets/sintel_384x768.py',
    '../_base_/schedules/schedule_s_fine.py', '../_base_/default_runtime.py'
]

# Train on FlyingChairs and finetune on Sintel
load_from = 'https://download.openmmlab.com/mmflow/flownet/flownetc_8x1_slong_flyingchairs_384x448.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.

from typing import Tuple

import torch
from torch import Tensor


def gradient(data: Tensor, stride: int = 1) -> Tuple[Tensor]:
    """Calculate gradient in data.

    Args:
        data (Tensor): input data with shape (B, C, H, W).
        stride (int): stride for distance of calculating changing. Default to
            1.

    Returns:
        tuple(Tensor): partial derivative of data with respect to x, with shape
            (B, C, H-stride, W), and partial derivative of data with respect to
            y with shape (B, C, H, W-stride).
    """
    D_dy = data[:, :, stride:] - data[:, :, :-stride]
    D_dx = data[:, :, :, stride:] - data[:, :, :, :-stride]
    return D_dx / stride, D_dy / stride


def smooth_1st_loss(flow: Tensor,
                    image: Tensor,
                    alpha: float = 0.,
                    smooth_edge_weighting: str = 'exponential') -> Tensor:
    """The First order smoothness loss.

    Modified from
    https://github.com/lliuz/ARFlow/blob/master/losses/flow_loss.py
    licensed under MIT License,
    and https://github.com/google-research/google-research/blob/master/uflow/uflow_utils.py
    licensed under the Apache License, Version 2.0.

    Args:
        flow (Tensor): Input optical flow with shape (B, 2, H, W).
        image (Tensor): Input image with shape (B, 3, H, W).
        alpha (float): Weight for modulates edge weighting. Default to: 0.
        smooth_edge_weighting (str): Function for calculating abstract
            value of image gradient which can be a string {'exponential'
            'gaussian'}.

    Returns:
        Tensor: A scaler of the first order smoothness loss.
    """ # noqa E501
    assert smooth_edge_weighting in ('exponential', 'gaussian'), (
        'smooth edge function must be `exponential` or `gaussian`,'
        f'but got {smooth_edge_weighting}')
    # Compute image gradients and sum them up to match the receptive field
    img_dx, img_dy = gradient(image)

    abs_fn = None
    if smooth_edge_weighting == 'gaussian':
        abs_fn = torch.square
    elif smooth_edge_weighting == 'exponential':
        abs_fn = torch.abs

    weights_x = torch.exp(-torch.mean(abs_fn(img_dx * alpha), 1, keepdim=True))
    weights_y = torch.exp(-torch.mean(abs_fn(img_dy * alpha), 1, keepdim=True))

    dx, dy = gradient(flow)

    loss_x = weights_x * dx.abs() / 2.
    loss_y = weights_y * dy.abs() / 2

    return loss_x.mean() + loss_y.mean()


def smooth_2nd_loss(flow: Tensor,
                    image: Tensor,
                    alpha: float = 0.,
                    smooth_edge_weighting: str = 'exponential'):
    """The Second order smoothness loss.

    Modified from
    https://github.com/lliuz/ARFlow/blob/master/losses/flow_loss.py
    licensed under MIT License,
    and https://github.com/google-research/google-research/blob/master/uflow/uflow_utils.py
    licensed under the Apache License, Version 2.0.

    Args:
        flow (Tensor): Input optical flow with shape (B, 2, H, W).
        image (Tensor): Input image with shape (B, 3, H, W).
        alpha (float): Weight for modulates edge weighting. Default to: 0.
        smooth_edge_weighting (str): Function for calculating abstract
            value of image gradient which can be a string {'exponential'
            'gaussian'}.

    Returns:
        Tensor: A scaler of the first order smoothness loss.
    """ # noqa E501

    assert smooth_edge_weighting in ('exponential', 'gaussian'), (
        'smooth edge function must be `exponential` or `gaussian`,'
        f'but got {smooth_edge_weighting}')
    # Compute image gradients and sum them up to match the receptive field
    img_dx, img_dy = gradient(image, stride=2)

    abs_fn = None
    if smooth_edge_weighting == 'gaussian':
        abs_fn = torch.square
    elif smooth_edge_weighting == 'exponential':
        abs_fn = torch.abs

    # weights_x with shape (B, 1, H, W-2)
    # weights_y with shape (B, 1, H-2, W)
    weights_x = torch.exp(-torch.mean(abs_fn(img_dx * alpha), 1, keepdim=True))
    weights_y = torch.exp(-torch.mean(abs_fn(img_dy * alpha), 1, keepdim=True))

    dx, dy = gradient(flow)
    dx2, _ = gradient(dx)  # (B, 1, H, W-2)
    _, dy2 = gradient(dy)  # (B, 1, H-2, W)

    loss_x = weights_x * dx2.abs() / 2.
    loss_y = weights_y * dy2.abs() / 2.

    return loss_x.mean() + loss_y.mean()



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence, Union

import torch
import torch.nn as nn
from mmcv.cnn import ConvModule
from mmengine.model import BaseModule


class DenseLayer(BaseModule):
    """Densely connected layer.

    Args:
        in_channels (int): Input channels of convolution module.
        feat_channels (int): Output channel of convolution module.
        conv_cfg (dict, optional): Config of convolution layer in module.
            Default: None.
        norm_cfg (dict, optional): Config of norm layer in module.
            Default: None.
        act_cfg (dict): Config of activation layer in module.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        init_cfg (dict, list, optional): Config dict of initialization of
            BaseModule. Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: int,
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 init_cfg: Optional[Union[list, dict]] = None) -> None:
        super().__init__(init_cfg=init_cfg)
        self.layers = ConvModule(
            in_channels=in_channels,
            out_channels=feat_channels,
            kernel_size=3,
            padding=1,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function for DenseLayer.

        Args:
            x (Tensor): The input feature.

        Returns:
            Tensor: The output feature of DenseLayer.
        """
        out = self.layers(x)
        return torch.cat((out, x), dim=1)


class BasicDenseBlock(BaseModule):
    """Basic Dense Block.

    A basic block which consists of several dense layers.

    Args:
        in_channels (int): Input channels of the block.
        feat_channels (Sequence[int]): Output channels of convolution module
            in dense layers. Default: (128, 128, 96, 64, 32).
        conv_cfg (dict, optional): Config of convolution layer in dense layers.
            Default: None.
        norm_cfg (dict, optional): Config of norm layer in dense layers.
            Default: None.
        act_cfg (dict, optional): Config of activation layer in dense layers.
            Default: None.
        init_cfg (dict, list, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 in_channels: int,
                 feat_channels: Sequence[int] = (128, 128, 96, 64, 32),
                 conv_cfg: Optional[dict] = None,
                 norm_cfg: Optional[dict] = None,
                 act_cfg: Optional[dict] = None,
                 init_cfg: Optional[Union[list, dict]] = None):
        super().__init__(init_cfg=init_cfg)
        self.in_channels = in_channels
        self.feat_channels = feat_channels
        layers = []
        for _feat_channels in feat_channels:
            layers.append(
                DenseLayer(in_channels, _feat_channels, conv_cfg, norm_cfg,
                           act_cfg))
            in_channels += _feat_channels
        self.layers = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward function for BasicDenseBlock.

        Args:
            x (Tensor): The input feature.

        Returns:
            Tensor: The output feature of BasicDenseBlock.
        """
        return self.layers(x)



# Copyright (c) OpenMMLab. All rights reserved.
import os
import os.path as osp
import re
from typing import Callable, List, Optional, Sequence, Union

from mmengine.dataset import BaseDataset as MME_BaseDataset

from mmflow.registry import DATASETS
from .utils import get_data_filename, load_ann_info, load_img_info

# these files contain nan, so exclude them.
exclude_files = dict(
    left_into_future=[
        '0004573.flo',
        '0006336.flo',
        '0016948.flo',
        '0015148.flo',
        '0006922.flo',
        '0003147.flo',
        '0003149.flo',
        '0000879.flo',
        '0006337.flo',
        '0014658.flo',
        '0015748.flo',
        '0001717.flo',
        '0000119.flo',
        '0017578.flo',
        '0004118.flo',
        '0004117.flo',
        '0004304.flo',
        '0004154.flo',
        '0011530.flo',
    ],
    right_into_future=[
        '0006336.flo',
        '0003148.flo',
        '0004117.flo',
        '0003666.flo',
    ],
    left_into_past=[
        '0000162.flo',
        '0004705.flo',
        '0006878.flo',
        '0004876.flo',
        '0004045.flo',
        '0000053.flo',
        '0005055.flo',
        '0000163.flo',
        '0000161.flo',
        '0000121.flo',
        '0000931.flo',
        '0005054.flo',
    ],
    right_into_past=[
        '0006878.flo',
        '0003147.flo',
        '0001549.flo',
        '0000053.flo',
        '0005034.flo',
        '0003148.flo',
        '0005055.flo',
        '0000161.flo',
        '0001648.flo',
        '0000160.flo',
        '0005054.flo',
    ])


@DATASETS.register_module()
class FlyingThings3DSubset(MME_BaseDataset):
    """FlyingThings3D subset dataset.

    Args:
        scene (str, optional): Scene in Flyingthings3D subset dataset,
            which has 3 options None, 'left' and 'right', and None means
            ``data_list`` will not be filter by scene filed.
            Defaults to None.
        ann_file (str): Annotation file path. Defaults to ''.
        metainfo (dict, optional): Meta information for dataset, such as class
            information. Defaults to None.
        data_root (str, optional): The root directory for ``data_prefix`` and
            ``ann_file``. Defaults to None.
        data_prefix (dict, optional): Prefix for training data. Defaults to
            dict(img=None, ann=None).
        filter_cfg (dict, optional): Config for filter data. Defaults to None.
        indices (int or Sequence[int], optional): Support using first few
            data in annotation file to facilitate training/testing on a smaller
            dataset. Defaults to None which means using all ``data_infos``.
        serialize_data (bool, optional): Whether to hold memory using
            serialized objects, when enabled, data loader workers can use
            shared RAM from master process instead of making a copy. Defaults
            to True.
        pipeline (list, optional): Processing pipeline. Defaults to [].
        test_mode (bool, optional): ``test_mode=True`` means in test phase.
            Defaults to False.
        lazy_init (bool, optional): Whether to load annotation during
            instantiation. In some cases, such as visualization, only the meta
            information of the dataset is needed, which is not necessary to
            load annotation file. ``Basedataset`` can skip load annotations to
            save time by set ``lazy_init=False``. Defaults to False.
        max_refetch (int, optional): If ``Basedataset.prepare_data`` get a
            None img. The maximum extra number of cycles to get a valid
            image. Defaults to 1000.
    """
    METAINFO = dict()

    def __init__(self,
                 scene: Optional[str] = None,
                 ann_file: str = '',
                 metainfo: Optional[dict] = None,
                 data_root: Optional[str] = None,
                 data_prefix: dict = dict(),
                 filter_cfg: Optional[dict] = None,
                 indices: Optional[Union[int, Sequence[int]]] = None,
                 serialize_data: bool = True,
                 pipeline: List[Union[dict, Callable]] = [],
                 test_mode: bool = False,
                 lazy_init: bool = False,
                 max_refetch: int = 1000) -> None:

        if scene is not None:
            assert scene in (
                'left', 'right'
            ), f'`scene` expected to be \'left\' or \'right\' but got {scene}'
        self.scene = scene

        super().__init__(ann_file, metainfo, data_root, data_prefix,
                         filter_cfg, indices, serialize_data, pipeline,
                         test_mode, lazy_init, max_refetch)

    def load_data_list(self) -> List[dict]:
        """Load ``data_list``

        ``data_list`` can be load from an annotation file named as
        ``self.ann_file`` or by parsing dataset path.

        Returns:
            list[dict]: A list of annotation.
        """
        if self.ann_file.endswith('json'):
            # load data_list with annotation file
            return super().load_data_list()
        else:
            # load data_list by path parsing
            self.load_data_info()
            return self.data_list

    def filter_data(self) -> List[dict]:
        """Filter data_list according to  ``scene``

        Returns:
            list[int]: Filtered results.
        """
        if self.scene is None:
            return self.data_list
        else:

            return [
                data_info for data_info in self.data_list
                if data_info['scene'] == self.scene
            ]

    def _get_data_dir(self):
        """Get the paths for images and optical flow."""

        self.flow_fw_dir = 'into_future'
        self.flow_bw_dir = 'into_past'
        self.flow_suffix = '.flo'
        self.occ_suffix = '.png'
        self.img1_suffix = '.png'
        self.img2_suffix = '.png'

        self.subset_dir = 'val' if self.test_mode else 'train'

        self.data_root = osp.join(self.data_root, self.subset_dir)
        img_root = osp.join(self.data_root, 'image_clean')
        flow_root = osp.join(self.data_root, 'flow')
        occ_root = osp.join(self.data_root, 'flow_occlusions')

        all_scene = os.listdir(img_root)

        self.img1_dir = [osp.join(img_root, s) for s in all_scene]
        self.img2_dir = [osp.join(img_root, s) for s in all_scene]
        self.flow_dir = [osp.join(flow_root, s) for s in all_scene]
        self.occ_dir = [osp.join(occ_root, s) for s in all_scene]

        self.flow_fw_dir = [osp.join(d, 'into_future') for d in self.flow_dir]
        self.flow_bw_dir = [osp.join(d, 'into_past') for d in self.flow_dir]
        self.occ_fw_dir = [osp.join(d, 'into_future') for d in self.occ_dir]
        self.occ_bw_dir = [osp.join(d, 'into_past') for d in self.occ_dir]

    def load_data_info(self) -> None:
        """Load data information, including file path of image1, image2 and
        optical flow."""

        self._get_data_dir()

        img1_filenames = []
        img2_filenames = []
        flow_fw_filenames = []
        flow_bw_filenames = []
        tmp_flow_fw_filenames = []
        tmp_flow_bw_filenames = []
        occ_fw_filenames = []
        occ_bw_filenames = []
        for _flow_fw_dir, _flow_bw_dir in zip(self.flow_fw_dir,
                                              self.flow_bw_dir):
            scene = _flow_fw_dir.split(os.sep)[-2]

            exc_key_fw = scene + '_into_future'
            exc_key_bw = scene + '_into_past'

            flow_fw_filenames_ = get_data_filename(_flow_fw_dir, None,
                                                   exclude_files[exc_key_fw])
            flow_bw_filenames_ = get_data_filename(_flow_bw_dir, None,
                                                   exclude_files[exc_key_bw])
            tmp_flow_fw_filenames += flow_fw_filenames_
            tmp_flow_bw_filenames += flow_bw_filenames_

        for ii in range(len(tmp_flow_fw_filenames)):
            flo_fw = tmp_flow_fw_filenames[ii]
            img1, img2, occ_fw, occ_bw, flo_bw = self._revise_dir(flo_fw)

            if (not (flo_bw in tmp_flow_bw_filenames) or not osp.isfile(img1)
                    or not osp.isfile(img2) or not osp.isfile(occ_fw)
                    or not osp.isfile(occ_bw)):
                continue
            img1_filenames.append(img1)
            img2_filenames.append(img2)
            flow_fw_filenames.append(flo_fw)
            flow_bw_filenames.append(flo_bw)
            occ_fw_filenames.append(occ_fw)
            occ_bw_filenames.append(occ_bw)

        load_img_info(self.data_list, img1_filenames, img2_filenames)
        load_ann_info(self.data_list, flow_fw_filenames, 'flow_fw_path')
        load_ann_info(self.data_list, occ_fw_filenames, 'occ_fw_path')
        load_ann_info(self.data_list, flow_bw_filenames, 'flow_bw_path')
        load_ann_info(self.data_list, occ_bw_filenames, 'occ_bw_path')

        for i in range(len(self.data_list)):
            self.data_list[i]['scene'] = self.data_list[i]['img1_path'].split(
                osp.sep)[-2]

    def _revise_dir(self, flow_fw_filename) -> Sequence[str]:
        """Revise directory of optical flow to get the directories of image and
        occlusion mask.

        Args:
            flow_fw_filename (str):  abstract file path of optical flow from
                image1 to image2.
        Returns:
            Sequence[str]: abstract paths of image1, image2, forward
                occlusion mask, backward occlusion mask and backward optical
                flow.
        """

        idx_f = int(osp.splitext(osp.basename(flow_fw_filename))[0])

        img1_filename = flow_fw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}image_clean{os.sep}')
        img1_filename = img1_filename.replace(f'{os.sep}into_future{os.sep}',
                                              f'{os.sep}')

        img1_filename = img1_filename.replace(self.flow_suffix,
                                              self.img1_suffix)
        img2_filename = re.sub(r'\d{7}', f'{idx_f+1:07d}', img1_filename)

        flow_bw_filename = flow_fw_filename.replace(
            f'{os.sep}into_future{os.sep}', f'{os.sep}into_past{os.sep}')
        flow_bw_filename = re.sub(r'\d{7}', f'{idx_f+1:07d}', flow_bw_filename)

        occ_fw_filename = flow_fw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}flow_occlusions{os.sep}')

        occ_fw_filename = occ_fw_filename.replace(self.flow_suffix,
                                                  self.occ_suffix)
        occ_bw_filename = flow_bw_filename.replace(
            f'{os.sep}flow{os.sep}', f'{os.sep}flow_occlusions{os.sep}')

        occ_bw_filename = occ_bw_filename.replace(self.flow_suffix,
                                                  self.occ_suffix)

        return (img1_filename, img2_filename, occ_fw_filename, occ_bw_filename,
                flow_bw_filename)



# Copyright (c) OpenMMLab. All rights reserved.
import numpy as np


def end_point_error_map(flow_pred: np.ndarray,
                        flow_gt: np.ndarray) -> np.ndarray:
    """Calculate end point error map.

    Args:
        flow_pred (ndarray): The predicted optical flow with the
            shape (H, W, 2).
        flow_gt (ndarray): The ground truth of optical flow with the shape
            (H, W, 2).

    Returns:
        ndarray: End point error map with the shape (H , W).
    """
    return np.sqrt(np.sum((flow_pred - flow_gt)**2, axis=-1))



# Copyright (c) OpenMMLab. All rights reserved.
import mmcv
import numpy as np


def adjust_hue(img, hue_factor):
    """Adjust hue of an image.

    Args:

        img (ndarray): Image to be adjust with BGR order.

        value (float): the amount of shift in H channel and must be in the
            interval [-0.5, 0.5].. 0.5 and -0.5 give complete reversal of hue
            channel in HSV space in positive and negative direction
            respectively. 0 means no shift. Therefore, both -0.5 and 0.5 will
            give an image with complementary colors while 0 gives the original
            image.

    Returns:
        ndarray: The hue-adjusted image.
    """
    if hue_factor is None:
        return img
    else:
        assert hue_factor >= -0.5 and hue_factor <= 0.5
        img = mmcv.bgr2hsv(img)
        img[:, :, 0] = (img[:, :, 0].astype(np.int) + int(hue_factor * 180.) +
                        180) % 180
        img = mmcv.hsv2bgr(img)
        return img


def adjust_gamma(img, gamma=1.0):
    """Using gamma correction to process the image.

    Args:
        img (ndarray): Image to be adjusted. uint8 datatype.

        gamma (float or int): Gamma value used in gamma correction. gamma is a
            positive value. Note: gamma larger than 1 make the shadows darker,
            while gamma smaller than 1 make dark regions lighter. Default: 1.0.
    """

    assert isinstance(gamma, float) or isinstance(gamma, int)
    assert gamma > 0

    assert img.dtype == 'uint8'

    table = ((np.arange(256) / 255.) ** gamma * (255 + 1 - 1e-3))\
        .astype('uint8')

    adjusted_img = mmcv.lut_transform(np.array(img, dtype=np.uint8), table)

    return adjusted_img



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Sequence, Tuple

import torch
from numpy import ndarray
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptSampleList, SampleList, TensorDict
from ..builder import build_flow_estimator
from ..decoders.maskflownet_decoder import Upsample
from ..utils import Warp
from .pwcnet import PWCNet


def centralize(img1: Tensor, img2: Tensor) -> Tuple[Tensor, Tensor]:
    """Centralize input images.

    Args:
        img1 (Tensor): The first input image.
        img2 (Tensor): The second input image.

    Returns:
        Tuple[Tensor, Tensor]: The first centralized image and the second
            centralized image.
    """
    rgb_mean = torch.cat((img1, img2), 2)
    rgb_mean = rgb_mean.view(rgb_mean.shape[0], 3, -1).mean(2)
    rgb_mean = rgb_mean.view(rgb_mean.shape[0], 3, 1, 1)
    return img1 - rgb_mean, img2 - rgb_mean, rgb_mean


@MODELS.register_module()
class MaskFlowNetS(PWCNet):
    """MaskFlowNetS model."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def extract_feat(self, imgs: Tensor) -> Sequence[TensorDict]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[TensorDict, TensorDict]: The feature pyramid of
                the first input image and the feature pyramid of seconde input
                image.
        """
        in_channels = self.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]
        img1, img2, _ = centralize(img1, img2)
        return self.encoder(img1), self.encoder(img2)


@MODELS.register_module()
class MaskFlowNet(MaskFlowNetS):
    """MaskFlowNet model."""

    def __init__(self,
                 *args,
                 maskflownetS: dict,
                 out_level: str = 'level2',
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.maskflownetS = build_flow_estimator(maskflownetS)
        self.out_level = out_level
        self.flow_div = self.decoder.flow_div

    def extract_feat(self, imgs: Tensor) -> Sequence[TensorDict]:
        """Extract features from images.

        Args:
            imgs (Tensor): The concatenated input images.

        Returns:
            Tuple[TensorDict, TensorDict, TensorDict,
                TensorDict, TensorDict]: The feature pyramid of
                the first input image and the feature pyramid of secode input
                image in stage1 and stage2 of MaskFlownet, and estimated
                multi-level flow from the stage1.
        """
        in_channels = self.maskflownetS.encoder.in_channels
        img1 = imgs[:, :in_channels, ...]
        img2 = imgs[:, in_channels:, ...]
        img1, img2, _ = centralize(img1, img2)

        feat1, feat2 = self.maskflownetS.extract_feat(imgs)
        flows_stage1, mask_stage1 = self.maskflownetS.decoder(feat1, feat2)

        img1 = torch.cat((img1, torch.zeros_like(mask_stage1)), dim=1)
        warped_img2 = Warp(align_corners=True)(
            img2, self.flow_div * Upsample(flows_stage1[self.out_level], 4))
        img2 = torch.cat((warped_img2, mask_stage1), dim=1)

        return feat1, feat2, self.encoder(img1), self.encoder(
            img2), flows_stage1

    def loss(self, imgs: Tensor, data_samples: SampleList) -> TensorDict:
        """Forward function for PWCNet when model training.

        Args:
            imgs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            TensorDict: The losses of output.
        """
        feat1, feat2, feat3, feat4, flows_stage1 = self.extract_feat(imgs)
        return self.decoder.loss(
            feat1=feat1,
            feat2=feat2,
            feat3=feat3,
            feat4=feat4,
            flows_stage1=flows_stage1,
            data_samples=data_samples)

    def predict(
            self,
            imgs: Tensor,
            data_samples: OptSampleList = None
    ) -> Sequence[Dict[str, ndarray]]:
        """Forward function for PWCNet when model testing.

        Args:
            imgs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        feat1, feat2, feat3, feat4, flows_stage1 = self.extract_feat(imgs)
        return self.decoder.predict(feat1, feat2, feat3, feat4, flows_stage1,
                                    data_samples)



_base_ = [
    '../_base_/models/raft.py',
    '../_base_/datasets/flyingthings3d_raft_400x720.py',
    '../_base_/schedules/raft_100k.py', '../_base_/default_runtime.py'
]

model = dict(freeze_bn=True, test_cfg=dict(iters=32))

optim_wrapper = dict(optimizer=dict(lr=0.000125, weight_decay=0.00001))
param_scheduler = dict(eta_max=0.000125)
val_cfg = dict(type='MultiValLoop')
test_cfg = dict(type='MultiTestLoop')
# Train on FlyingChairs and finetune on FlyingThings3D
load_from = 'https://download.openmmlab.com/mmflow/raft/raft_8x2_100k_flyingchairs.pth'  # noqa



MaskflownetS_checkpoint = 'https://download.openmmlab.com/mmflow/maskflownet/maskflownets_8x1_sfine_flyingthings3d_subset_384x768.pth'  # noqa
model = dict(
    type='MaskFlowNet',
    data_preprocessor=dict(
        type='FlowDataPreprocessor',
        mean=[0., 0., 0.],
        std=[255., 255., 255.],
        bgr_to_rgb=False,
        sigma_range=(0, 0.04),
        clamp_range=(0., 1.)),
    maskflownetS=dict(
        type='MaskFlowNetS',
        freeze_net=True,
        encoder=dict(
            type='PWCNetEncoder',
            in_channels=3,
            net_type='Basic',
            pyramid_levels=[
                'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
            ],
            out_channels=(16, 32, 64, 96, 128, 196),
            strides=(2, 2, 2, 2, 2, 2),
            dilations=(1, 1, 1, 1, 1, 1),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
        decoder=dict(
            type='MaskFlowNetSDecoder',
            warp_in_channels=dict(
                level6=196, level5=128, level4=96, level3=64, level2=32),
            up_channels=dict(
                level6=16, level5=16, level4=16, level3=16, level2=16),
            warp_type='AsymOFMM',
            in_channels=dict(
                level6=81, level5=227, level4=195, level3=163, level2=131),
            corr_cfg=dict(type='Correlation', max_displacement=4),
            act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
            scaled=False,
            post_processor=dict(type='ContextNet', in_channels=579)),
        # model training and testing settings
        train_cfg=dict(),
        test_cfg=dict(),
        init_cfg=dict(type='Pretrained', checkpoint=MaskflownetS_checkpoint)),
    encoder=dict(
        type='PWCNetEncoder',
        in_channels=4,
        net_type='Basic',
        pyramid_levels=[
            'level1', 'level2', 'level3', 'level4', 'level5', 'level6'
        ],
        out_channels=(16, 32, 64, 96, 128, 196),
        strides=(2, 2, 2, 2, 2, 2),
        dilations=(1, 1, 1, 1, 1, 1),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1)),
    decoder=dict(
        type='MaskFlowNetDecoder',
        warp_in_channels=dict(
            level6=196, level5=128, level4=96, level3=64, level2=32),
        up_channels=dict(
            level6=16, level5=16, level4=16, level3=16, level2=16),
        warp_type='Basic',
        with_mask=False,
        in_channels=dict(
            level6=52, level5=198, level4=166, level3=134, level2=102),
        corr_cfg=dict(type='Correlation', max_displacement=2),
        act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
        scaled=False,
        post_processor=dict(type='ContextNet', in_channels=550),
        flow_loss=dict(
            type='MultiLevelEPE',
            p=2,
            reduction='sum',
            weights={
                'level2': 0.005,
                'level3': 0.01,
                'level4': 0.02,
                'level5': 0.08,
                'level6': 0.32
            }),
    ),
    # model training and testing settings
    train_cfg=dict(),
    test_cfg=dict(),
    init_cfg=dict(
        type='Kaiming', a=0.1, distribution='uniform', layer='Conv2d'))
randomness = dict(seed=0, diff_rank_seed=True)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import build_activation_layer
from mmcv.ops import DeformConv2d
from mmengine.model import BaseModule
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptMultiConfig, OptSampleList, SampleList, TensorDict
from ..utils import CorrBlock
from .pwcnet_decoder import PWCModule, PWCNetDecoder


def Upsample(img, factor) -> Tensor:
    """Upsampling function.

    Args:
        img (tensor(B, C, H, W)): Input image for upsampling.
        factor (int): Upsampling factor.

    Returns:
        Upsampled image.
    """
    if factor == 1:
        return img
    _, _, H, W = img.shape
    img = F.pad(img, [0, 1, 0, 1], mode='replicate')
    upsample_img = F.interpolate(
        img, (H * factor + 1, W * factor + 1),
        mode='bilinear',
        align_corners=True)
    upsample_img = upsample_img[:, :, :-1, :-1]

    return upsample_img


class BasicDeformWarpBlock(BaseModule):
    """Basic Deformable Warp Block.

    Args:
        channels (int): Input channels of deformable convolution.
        act_cfg (dict): Config dict of activation layer.
        with_deform_bias (bool): Whether to use bias in deformable convolution
            or not. Default: True.
        init_cfg (dict, optional): Config for module initialization.
            Default: None.
    """

    def __init__(self,
                 channels: int,
                 act_cfg: dict,
                 with_deform_bias: bool = True,
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(init_cfg)
        self.channels = channels
        self.deconv = DeformConv2d(channels, channels, 3, padding=1)
        self.act = build_activation_layer(act_cfg)
        self.with_deform_bias = with_deform_bias
        if self.with_deform_bias:
            self.deconv_bias = nn.Parameter(torch.zeros(channels, 1, 1))

    def forward(self, feat2: Tensor, flow: Tensor) -> Tensor:
        """Forward function for BasicDeformWarpBlock.

        Args:
            feat2 (Tensor): The feature from the second image.
            flow (Tensor): The optical flow used for DCN offset.

        Returns:
            Tensor: The output feature from BasicDeformWarpBlock.
        """
        B, _, H, W = flow.shape
        conv_offset = torch.repeat_interleave(
            flow[:, None], repeats=9, dim=1).reshape(B, -1, H, W)
        if self.with_deform_bias:
            deform_feat = self.deconv(
                feat2, conv_offset) + self.deconv_bias.expand(B, -1, -1, -1)
        else:
            deform_feat = self.deconv(feat2, conv_offset)
        return self.act(deform_feat)


class DeformWarpBlock(BaseModule):
    """Deformable Warp Block.

    Args:
        channels (int): Input channels of deformable convolution.
        up_channels (int): Input channels of trade-off convolution layer for
            upsampled feat.
        act_cfg (dict): Config of activation layer.
        with_deform_bias (bool, optional): Whether to use bias in
            deformable convolution or not. Default: True.
        init_cfg (dict, optional): Config dict for initialization of module.
            Default: None.
    """

    def __init__(self,
                 channels: int,
                 up_channels: int,
                 act_cfg: dict,
                 with_deform_bias: bool = True,
                 init_cfg: OptMultiConfig = None) -> None:
        super().__init__(init_cfg)
        self.channels = channels
        self.deconv = DeformConv2d(channels, channels, 3, padding=1)
        self.tradeoff_conv = nn.Conv2d(up_channels, channels, 3, padding=1)
        self.act = build_activation_layer(act_cfg)
        self.with_deform_bias = with_deform_bias
        if self.with_deform_bias:
            self.deconv_bias = nn.Parameter(torch.zeros(channels, 1, 1))

    def forward(self, feat2: Tensor, flow: Tensor, mask_feat: Tensor,
                up_feat: Tensor) -> Tensor:
        """Forward function for DeformWarpBlock.

        Args:
            feat2 (Tensor): The feature from the second image.
            flow (Tensor): The optical flow used for DCN offset.
            mask_feat (Tensor): The learnable occlusion mask.
            up_feat (Tensor): The upsampled feature from the last level.

        Returns:
            Tensor: The output feature from DeformWarpBlock.
        """

        assert mask_feat.shape[1] == 1
        B, _, H, W = flow.shape
        # repeat = kernel_size*kernel_size
        conv_offset = torch.repeat_interleave(
            flow[:, None], repeats=9, dim=1).reshape(B, -1, H, W)
        if self.with_deform_bias:
            deform_feat = self.deconv(
                feat2, conv_offset) + self.deconv_bias.expand(B, -1, -1, -1)
        else:
            deform_feat = self.deconv(feat2, conv_offset)
        tradeoff_feat = self.tradeoff_conv(up_feat)
        mask_feat = torch.squeeze(torch.sigmoid(mask_feat), dim=1)
        warp_feat = torch.einsum('ijkl, ikl->ijkl', deform_feat,
                                 mask_feat) + tradeoff_feat
        return self.act(warp_feat)


class WarpCorrBlock(BaseModule):
    """Warp Correlation Block.

    Args:
        channels (int): Input channels of DeformWarpBlock.
        corr_cfg (dict): Config of BasicCorrBlock.
        up_channels (int, optional): Input channels of trade-off convolution
            layer for upsampled feat, if set to None, the warp_type must be
            'Basic'. Default: None.
        warp_type (str): Type of Warp block, it has 2 options 'Basic' and
            'AsyOFMM'. Default: "Basic".
        act_cfg (dict): Config dict of activation layer.
            Default: dict(type='LeakyReLU', negative_slope=0.1).
        scaled (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not.
            Defaults to False.
        with_deform_bias (bool): Whether to use bias in DeformConv2d or not.
            Default: True.
        init_cfg (dict, optional): Config dict for initialization of module.
            Default: None.
    """

    def __init__(self,
                 channels: int,
                 corr_cfg: dict,
                 up_channels: Optional[int] = None,
                 warp_type: str = 'Basic',
                 act_cfg=dict(type='LeakyReLU', negative_slope=0.1),
                 scaled: bool = False,
                 with_deform_bias: bool = True,
                 init_cfg: OptMultiConfig = None) -> None:

        super().__init__(init_cfg=init_cfg)

        assert warp_type in ('Basic', 'AsymOFMM')

        self.channels = channels
        self.up_channels = up_channels
        self.corr_cfg = corr_cfg
        self.warp_type = warp_type
        self.act_cfg = act_cfg
        self.scaled = scaled
        self.with_deform_bias = with_deform_bias
        self._make_layers()

    def _make_layers(self) -> None:
        """Build warp and correlation block as the given type."""
        if self.warp_type == 'AsymOFMM':
            self.warp = DeformWarpBlock(
                channels=self.channels,
                up_channels=self.up_channels,
                act_cfg=self.act_cfg,
                with_deform_bias=self.with_deform_bias)
        elif self.warp_type == 'Basic':
            self.warp = BasicDeformWarpBlock(
                channels=self.channels,
                act_cfg=self.act_cfg,
                with_deform_bias=self.with_deform_bias)
        self.corr = CorrBlock(self.corr_cfg, self.act_cfg, scaled=self.scaled)

    def forward(self,
                feat1: Tensor,
                feat2: Tensor,
                up_flow: Tensor,
                up_mask: Optional[Tensor] = None,
                up_feat: Optional[Tensor] = None) -> Tensor:
        """Forward function for WarpCorrBlock.

        Args:
            feat1 (Tensor): The feature from the first input image.
            feat2 (Tensor): The feature from the second input image.
            up_flow (Tensor):  The upsampled optical flow from the last level.
            up_mask (Tensor, optional): The upsampled learnable occlusion mask
                from the last level. Defaults to None.
            up_feat (Tensor, optional): The upsampled feature from the last
                level.

        Returns:
            Tensor: The correlation between feat1 and the warped feat2.
        """
        if self.warp_type == 'AsymOFMM':
            warp_feat = self.warp(feat2, up_flow, up_mask, up_feat)
        elif self.warp_type == 'Basic':
            warp_feat = self.warp(feat2, up_flow)
        corr_feat = self.corr(feat1, warp_feat)

        return corr_feat


class MaskModule(PWCModule):
    """Basic module of Mask-FlowNet.

    Args:
        up_channels (int): Output channels of feature upsampling layer.
        with_mask (bool): Whether to predict mask or not.
    """

    def __init__(self, up_channels: int, with_mask: bool, *args,
                 **kwargs) -> torch.Tensor:
        self.up_channels = up_channels
        self.with_mask = with_mask
        super().__init__(*args, **kwargs)

    def _make_predict_layer(self) -> None:
        """Make prediction layer."""
        self.predict_flow = nn.Conv2d(
            self.last_channels, 2, kernel_size=3, padding=1)

        if self.with_mask:
            self.predict_mask = nn.Conv2d(
                self.last_channels, 1, kernel_size=3, padding=1)

    def _make_upsample_layer(self) -> None:
        """Make upsample layer."""
        if self.up_flow:
            self.upfeat_layer = nn.Sequential(
                nn.ConvTranspose2d(
                    self.last_channels,
                    out_channels=self.up_channels,
                    kernel_size=4,
                    stride=2,
                    padding=1), build_activation_layer(self.act_cfg))

    def forward(
        self, x: Tensor, upflow: Tensor
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        """Forward function for MaskModule.

        Args:
            x (Tensor): The input feature.
            upflow (Tensor): The upsampled optical flow from the last level.

        Returns:
            Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]: The
                predicted optical flow, the predicted occlusion mask,
                the feature to predict optical flow, the upsample flow from the
                last level, the upsampled occlusion mask from the last level,
                and upsampled feature from the last level.
        """
        feat = self.dense_net(x)
        flow = self.predict_flow(feat) + upflow

        mask = None
        upflow = None
        upmask = None
        upfeat = None
        if self.with_mask:
            mask = self.predict_mask(feat)
            if self.up_flow:
                upmask = Upsample(mask, factor=2)
        if self.up_flow:
            upflow = Upsample(flow, factor=2)
            # upfeat will used to make trade-off feat.
            upfeat = self.upfeat_layer(feat)

        return flow, mask, feat, upflow, upmask, upfeat


@MODELS.register_module()
class MaskFlowNetSDecoder(PWCNetDecoder):
    """The decoder module of MaskFlowNetS.

    Args:
        warp_in_channels (dict): Dict input channels of warp block.
        up_channels (dict): Dict input channels of feat upsampling layer at
            each level.
        with_mask (bool, optional): Whether to predict mask or not.
            If not define, decoder predicts mask depending on whether upsample
            feat from the last level. Default: None.
        warp_type (str): Type of Warp block, it has 2 options 'Basic' and
            'AsymOFMM'. Default: "Basic".
        with_deform_bias (bool): Whether to use bias in DeformConv2d or not.
            Default: True.
    """

    def __init__(self,
                 warp_in_channels: Dict[str, int],
                 up_channels: Dict[str, int],
                 with_mask: Optional[bool] = None,
                 warp_type: str = 'AsymOFMM',
                 with_deform_bias: bool = True,
                 *args,
                 **kwargs) -> None:

        assert isinstance(warp_in_channels, dict)
        assert isinstance(up_channels, dict)

        self.warp_in_channels = warp_in_channels
        self.up_channels = up_channels
        self.with_mask = with_mask
        self.warp_type = warp_type
        self.with_deform_bias = with_deform_bias

        super().__init__(*args, **kwargs)

    def _make_layers(self) -> None:
        """Build sub-modules of this decoder."""
        layers = []
        for level in self.flow_levels:
            up_sample = level != self.end_level
            # predicts mask depending on whether upsample
            with_mask = up_sample if self.with_mask is None else self.with_mask
            up_channels = self.up_channels.get(level)

            layers.append([
                level,
                self._make_layer(self.in_channels[level], up_channels,
                                 up_sample, with_mask)
            ])
            self.decoders = nn.ModuleDict(layers)

    def _make_layer(self,
                    in_channels: int,
                    up_channels: Optional[int] = None,
                    up_sample: bool = True,
                    with_mask: bool = True) -> torch.nn.Module:
        """Build module at each level of this decoder.

        Args:
            in_channels (int): The channels of input feature
            up_channels (int, optional): The channels of upsampled features.
                Defaults to None.
            up_sample (bool): Whether upsample feature for the next level.
                Defaults to True.
            with_mask (bool): Whether predict occlusion mask.
                Defaults to True.

        Returns:
            torch.nn.Module: The sub-module for this decoder.
        """
        return MaskModule(
            up_channels,
            with_mask,
            in_channels,
            up_sample,
            self.densefeat_channels,
            act_cfg=self.act_cfg)

    def _make_corr_block(self, corr_cfg: dict, act_cfg: dict,
                         scaled: bool) -> None:
        """Make correlation block for different optical flow.

        Args:
            corr_cfg (dict): Config for correlation layer.
            act_cfg (dict): Config of activation layer in module.
            scaled (bool): Whether to use scaled correlation by the number of
                elements involved to calculate correlation or not.
        """
        self.corr_block = nn.ModuleDict()

        for level in self.flow_levels:
            if level == self.start_level:
                self.corr_block[level] = CorrBlock(corr_cfg, act_cfg, scaled)
            else:
                self.corr_block[level] = WarpCorrBlock(
                    channels=self.warp_in_channels[level],
                    corr_cfg=corr_cfg,
                    up_channels=self.up_channels[level],
                    warp_type=self.warp_type,
                    act_cfg=act_cfg,
                    scaled=scaled,
                    with_deform_bias=self.with_deform_bias)

    def forward(
            self, feat1: TensorDict,
            feat2: TensorDict) -> Union[TensorDict, Tuple[TensorDict, Tensor]]:
        """Forward function for MaskFlowNet decoder.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.

        Returns:
            Dict[str, Tensor]: The predicted multi-levels optical flow.
        """
        minH, minW = feat1[self.start_level].shape[2:]
        flow_pred = dict()
        flow = None
        upmask = None
        upfeat = None
        last_mask = None
        upflow = torch.zeros(1, 2, minH, minW).to(feat1[self.start_level])

        for level in self.flow_levels[::-1]:
            _feat1, _feat2 = feat1[level], feat2[level]

            if level == self.start_level:
                corr_feat = self.corr_block[level](_feat1, _feat2)
            else:

                corr_feat_ = self.corr_block[level](
                    _feat1, _feat2, upflow * self.multiplier[level], upmask,
                    upfeat)
                corr_feat = torch.cat((corr_feat_, _feat1, upfeat, upflow),
                                      dim=1)
            flow, _, feat, upflow, upmask, upfeat = self.decoders[level](
                corr_feat, upflow)

            if level == 'level3':
                # the upsampled mask in level3 is the last mask in stage1
                # of MaskFlowNet.
                last_mask = upmask

            # offset in dcn is (y0,x0,...) but in value in flow map is (u, v)
            flow_pred[level] = flow.flip(1)

        if self.post_processor is not None:
            post_flow = self.post_processor(feat)
            flow_pred[self.end_level] = flow_pred[
                self.flow_levels[0]] + post_flow.flip(1)
        if last_mask is not None:
            last_mask = Upsample(last_mask, 4)
        # Stage2 of MaskFlowNet need input mask.
        return flow_pred, last_mask

    def predict(self,
                feat1: TensorDict,
                feat2: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model testing.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred, _ = self.forward(feat1, feat2)
        flow_results = flow_pred[self.end_level]
        return self.predict_by_feat(flow_results, data_samples)

    def loss(self, feat1: TensorDict, feat2: TensorDict,
             data_samples: SampleList) -> TensorDict:
        """Forward function when model training.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """

        flow_pred, _ = self.forward(feat1, feat2)
        return self.loss_by_feat(flow_pred, data_samples)


@MODELS.register_module()
class MaskFlowNetDecoder(MaskFlowNetSDecoder):
    """The decoder module of MaskFlowNet."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _make_corr_block(self, corr_cfg: dict, act_cfg: dict,
                         scaled: bool) -> None:
        self.corr_layer = CorrBlock(
            self.corr_cfg, self.act_cfg, scaled=self.scaled)
        self.corr_block = nn.ModuleDict()

        for level in self.flow_levels:
            self.corr_block[level] = WarpCorrBlock(
                channels=self.warp_in_channels[level],
                corr_cfg=corr_cfg,
                up_channels=self.up_channels[level],
                warp_type=self.warp_type,
                act_cfg=act_cfg,
                scaled=scaled,
                with_deform_bias=self.with_deform_bias)

    def forward(
        self,
        feat1: TensorDict,
        feat2: TensorDict,
        feat3: TensorDict,
        feat4: TensorDict,
        flows_stage1: TensorDict,
    ) -> TensorDict:
        """Forward function for MaskFlowNetDecoder.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image from stage1 of MaskFlowNet.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image from stage1 of MaskFlowNet.
            feat3 (Dict[str, Tensor]): The feature pyramid from the first
                image from stage2 of MaskFlowNet.
            feat4 (Dict[str, Tensor]): The feature pyramid from the second
                image from stage2 of MaskFlowNet.
            flows_stage1 (Dict[str, Tensor]): Estimated multi-level flow from
                the stage1.

        Returns:
            Dict[str, Tensor]: The predicted multi-levels optical flow.
        """

        flows_pred = dict()
        upfeat = None
        upflow = None

        for level in self.flow_levels[::-1]:
            i_feat1, i_feat2, i_feat3, i_feat4, i_flow = feat1[level], feat2[
                level], feat3[level], feat4[level], flows_stage1[level]

            if level == self.start_level:
                upflow = i_flow

            corr_feat1 = self.corr_block[level](
                i_feat1, i_feat2, self.multiplier[level] * upflow)
            corr_feat2 = self.corr_layer(i_feat3, i_feat4)
            corr_feat = torch.cat((corr_feat1, corr_feat2), dim=1)

            if upfeat is None:
                flow, _, feat, upflow, _, upfeat = self.decoders[level](
                    torch.cat((corr_feat, upflow), dim=1), upflow)
            else:
                flow, _, feat, upflow, _, upfeat = self.decoders[level](
                    torch.cat((i_feat1, upfeat, corr_feat, upflow, i_flow),
                              dim=1), upflow)

            flows_pred[level] = flow.flip(1)

        if self.post_processor is not None:
            post_flow = self.post_processor(feat)
            flows_pred[self.end_level] = flows_pred[
                self.flow_levels[0]] + post_flow.flip(1)

        return flows_pred

    def loss(self, feat1: TensorDict, feat2: TensorDict, feat3: TensorDict,
             feat4: TensorDict, flows_stage1: TensorDict,
             data_samples: SampleList) -> TensorDict:
        """Forward function when model training.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image from stage1 of MaskFlowNet.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image from stage1 of MaskFlowNet.
            feat3 (Dict[str, Tensor]): The feature pyramid from the first
                image from stage2 of MaskFlowNet.
            feat4 (Dict[str, Tensor]): The feature pyramid from the second
                image from stage2 of MaskFlowNet.
            flows_stage1 (Dict[str, Tensor]): Estimated multi-level flow from
                the stage1.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The dict of losses.
        """

        flow_pred = self.forward(feat1, feat2, feat3, feat4, flows_stage1)

        return self.loss_by_feat(flow_pred, data_samples)

    def predict(self,
                feat1: TensorDict,
                feat2: TensorDict,
                feat3: TensorDict,
                feat4: TensorDict,
                flows_stage1: TensorDict,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function when model testing.

        Args:
            feat1 (Dict[str, Tensor]): The feature pyramid from the first
                image.
            feat2 (Dict[str, Tensor]): The feature pyramid from the second
                image.
            feat3 (Dict[str, Tensor]): The feature pyramid from the first
                image from stage2 of MaskFlowNet.
            feat4 (Dict[str, Tensor]): The feature pyramid from the second
                image from stage2 of MaskFlowNet.
            flows_stage1 (Dict[str, Tensor]): Estimated multi-level flow from
                the stage1.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.
        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        flow_pred = self.forward(feat1, feat2, feat3, feat4, flows_stage1)
        flow_result = flow_pred[self.end_level]
        return self.predict_by_feat(flow_result, data_samples)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Sequence

import cv2
import numpy as np
from mmcv.transforms import BaseTransform
from mmcv.transforms.utils import cache_randomness

from mmflow.registry import TRANSFORMS

img_keys = ['img1', 'img2']

flow_keys = ['gt_flow_fw', 'gt_flow_bw']

occ_keys = ['gt_occ_fw', 'gt_occ_bw']

valid_keys = ['gt_valid_fw', 'gt_valid_bw']


def theta_is_valid(theta: np.ndarray) -> bool:
    """Whether affine transform theta is a valid affine transform.

    A valid affine transform is an affine transform which guarantees the
    transformed image covers the whole original picture frame.

    Args:
        theta (ndarray): affine transform matrix.
    Returns:
        bool: whether this transform matrix is valid.
    """

    bounds = np.array([
        [-0.5, -0.5, 1.],  # left top
        [-0.5, 0.5, 1.],  # left bottom
        [0.5, -0.5, 1.],  # right top
        [0.5, 0.5, 1.],  # right bottom
    ])
    """
    (-0.5, -0.5)          (0.5, -0.5)
                 --------
                |        |
                |        |
                |        |
                 --------
    (-0.5, 0.5)          (0.5, 0.5)
    """
    bounds = (np.linalg.inv(theta) @ bounds.T).T

    valid = ((bounds[:, :2] >= -0.5) & (bounds[:, :2] <= 0.5)).all()

    return valid


def check_out_of_bound(flow: np.ndarray, occ: np.ndarray) -> np.ndarray:
    """Check pixels that will move out of bound after warping by flow and mark
    as occluded pixels.

    Revise occlusion mask for transformed optical flow data.

    Args:
        flow (ndarray): optical flow data.
        occ (ndarray): original occlusion mask.
    Returns:
        ndarray: the occlusion mask for optical flow.
    """

    height, width, _ = flow.shape

    xx, yy = np.meshgrid(range(width), range(height))

    xx = xx.astype(flow.dtype)
    yy = yy.astype(flow.dtype)

    xx += flow[:, :, 0]
    yy += flow[:, :, 1]

    out_of_bound = ((xx < 0) | (yy < 0) | (xx >= width) |
                    (yy >= height)).astype(occ.dtype)

    occ = np.clip(out_of_bound + occ, 0, 1)

    return occ


def transform_img(img: np.ndarray, theta: np.ndarray, height: int,
                  width: int) -> np.ndarray:
    """Transform image with cv2 warpAffine.

    Args:
        img (ndarray): image that will be transformed.
        theta (ndarray): transform matrix.
        height (int): height of output image.
        width (int): width of output image.

    Returns:
        ndarray: transformed image.
    """
    return cv2.warpAffine(img, theta[:2, :], (width, height))


def transform_flow(flow: np.ndarray, valid: np.ndarray, theta1: np.ndarray,
                   theta2: np.ndarray, height: int, width: int) -> np.ndarray:
    """Transform optical flow with cv2 warpAffine.

    Args:
        flow (ndarray): flow that will be transformed.
        theta1 (ndarray): global transform matrix.
        theta2 (ndarray): relative transform matrix.
        height (int): height of output image.
        width (int): width of output image.

    Returns:
        ndarray: transformed optical flow.
    """

    flow_ = cv2.warpAffine(flow, theta1[:2, :], (width, height))
    if valid is not None:
        flow_ = flow_ / (valid[:, :, None] + 1e-12)
    """
    X1                 Affine(theta1)             X1'
               x                                   x
    theta1(-1) y           ->                      y
               1                                   1

    X2                 Affine(theta2)             X2'
               x   u                                         x   u
    theta1(-1) y + v       ->           theta2 x {theta1(-1) y + v}
               1   0                                         1   0
                                        flow' = X2' -X1'
    """

    # (u, v) -> (u, v, 0); shape (height, width, 2) -> (height, width, 3)
    homo_flow_ = np.concatenate((flow_, np.zeros((height, width, 1))), axis=2)

    xx, yy = np.meshgrid(range(width), range(height))

    # grid of homogeneous coordinates
    homo_grid = np.stack((xx, yy, np.ones((height, width))),
                         axis=2).astype(flow.dtype)
    """
    theta2 x [u, v, 0]T + (theta2 x theta1(-1) - [1, 1, 1]) x [x, y, 1]T
    """
    flow_final = homo_grid @ (theta2 @ np.linalg.inv(theta1) -
                              np.eye(3)).T + homo_flow_ @ theta2.T

    return flow_final[:, :, :2]


@TRANSFORMS.register_module()
class RandomAffine(BaseTransform):
    """Random affine transformation of images, flow map and occlusion map (if
    available).

    Required Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)


    Modified Keys:

    - img1
    - img2
    - gt_flow_fw (optional)
    - gt_flow_bw (optional)
    - gt_occ_fw (optional)
    - gt_occ_bw (optional)
    - gt_valid_fw (optional)
    - gt_valid_bw (optional)

    Added Keys:

    - global_ndc_affine_mat
    - relative_ndc_affine_mat

    Keys of global_transform and relative_transform should be the subset of
    ('translates', 'zoom', 'shear', 'rotate'). And also, each key and its
    corresponding values has to satisfy the following rules:
        - translates: the translation ratios along x axis and y axis. Defaults
            to(0., 0.).
        - zoom: the min and max zoom ratios. Defaults to (1.0, 1.0).
        - shear: the min and max shear ratios. Defaults to (1.0, 1.0).
        - rotate: the min and max rotate degree. Defaults to (0., 0.).

    Args:
        global_transform (dict): A dict which contains keys: transform, zoom,
            shear, rotate. global_transform will transform both img1 and img2.
        relative_transform (dict): A dict which contains keys: transform, zoom,
            shear, rotate. relative_transform will only transform img2 after
            global_transform to both images.
        preserve_valid (bool): Whether continue transforming until both images
            are valid. A valid affine transform is an affine transform which
            guarantees the transformed image covers the whole original picture
            frame. Defaults to True.
        check_bound (bool): Whether to check out of bound for transformed
            occlusion maps. If True, all pixels in borders of img1 but not in
            borders of img2 will be marked occluded. Defaults to False.
    """

    def __init__(self,
                 global_transform: Optional[dict] = None,
                 relative_transform: Optional[dict] = None,
                 preserve_valid: bool = True,
                 check_bound: bool = False) -> None:

        self.DEFAULT_TRANSFORM = dict(
            translates=(0., 0.),
            zoom=(1.0, 1.0),
            shear=(1.0, 1.0),
            rotate=(0., 0.))

        self.global_transform = self._check_input(global_transform)
        self.relative_transform = self._check_input(relative_transform)

        assert isinstance(preserve_valid, bool)
        self.preserve_valid = preserve_valid

        assert isinstance(check_bound, bool)
        self.check_bound = check_bound

    def _check_input(self, transform: dict) -> dict:
        """Check whethere input transform.

        Args:
            transform (dict): A dict which may contains keys: transform, zoom,
                shear, rotate. If transform miss some key, it will be set the
                default value.

        Returns:
            dict: transform dict with all valid values.
        """

        ret = dict() if not isinstance(transform, dict) else transform.copy()

        assert set(ret).issubset(self.DEFAULT_TRANSFORM), (
            f'Got unexpected keys in {transform}. \n'
            f"Valid keys should be the subset of ('translates', 'zoom', "
            f"'shear', 'rotate')")

        for k in self.DEFAULT_TRANSFORM:
            if k not in ret:
                ret[k] = self.DEFAULT_TRANSFORM[k]

            assert isinstance(ret[k], (list, tuple))
            assert len(ret[k]) == 2
            assert ret[k][0] <= ret[k][1]

        return ret

    def transform(self, results: dict) -> dict:
        """

        Args:
            results (dict): data including image, annotation and meta
                information in data augmentation pipeline.

        Returns:
            dict: transformed data.
        """

        h, w = results['img_shape']

        # theta0_ndc, theta1_ndc and theta2_ndc are 3 x 3 affine transformation
        # matrix in normal device coordinates, with origin at the center of
        # pictures and picture's width range and height range from [-0.5, 0.5]
        # and [-0.5, 0.5].
        theta0_ndc = np.identity(3)

        # apply global transform to identity matrix theta0_ndc
        theta1_ndc = self._apply_random_affine_to_theta(
            theta0_ndc, **self.global_transform)

        # apply relative transform to theta1_ndc
        theta2_ndc = self._apply_random_affine_to_theta(
            theta1_ndc, **self.relative_transform)

        # T is similar transform matrix
        T = np.array([[1. / (w - 1.), 0., -0.5], [0., 1. / (h - 1.), -0.5],
                      [0., 0., 1.]], np.float32)

        T_inv = np.linalg.inv(T)

        # theta1_world and theta2_world are affine transformations in world
        # coordinates, with origin at top left corner of pictures and picture's
        # width range and height range from [0, width] and [0, height].
        theta1_world = T_inv @ theta1_ndc @ T
        theta2_world = T_inv @ theta2_ndc @ T
        theta_world_li = [theta1_world, theta2_world]

        # transform img1 and img2
        for i in range(len(img_keys)):
            results[img_keys[i]] = transform_img(results[img_keys[i]],
                                                 theta_world_li[i], h, w)
        flow_keys_in = []
        for k in flow_keys:
            if results.get(k, None) is not None:
                flow_keys_in.append(k)
        occ_keys_in = []
        for k in occ_keys:
            if results.get(k, None) is not None:
                occ_keys_in.append(k)
        valid_keys_in = []
        for k in valid_keys:
            if results.get(k, None) is not None:
                valid_keys_in.append(k)
        # transform flows and valid
        for i in range(len(flow_keys_in)):
            if len(valid_keys_in) == len(flow_keys_in):
                valid = results[valid_keys_in[i]]

                results[valid_keys[i]] = transform_img(results[valid_keys[i]],
                                                       theta_world_li[i], h, w)
                results[flow_keys_in[i]] = transform_flow(
                    flow=results[flow_keys_in[i]] * valid[:, :, None],
                    valid=results[valid_keys[i]],
                    theta1=theta_world_li[i],
                    theta2=theta_world_li[1 - i],
                    height=h,
                    width=w)

            else:
                results[flow_keys_in[i]] = transform_flow(
                    flow=results[flow_keys_in[i]],
                    valid=None,
                    theta1=theta_world_li[i],
                    theta2=theta_world_li[1 - i],
                    height=h,
                    width=w)

        # transform occlusion if available
        for i in range(len(occ_keys_in)):
            results[occ_keys_in[i]] = transform_img(results[occ_keys_in[i]],
                                                    theta_world_li[i], h, w)
            if self.check_bound:
                results[occ_keys[i]] = check_out_of_bound(
                    results[flow_keys_in[i]], results[occ_keys_in[i]])

        # create new meta 'global_ndc_affine_mat'
        results['global_ndc_affine_mat'] = theta1_ndc

        # create new meta 'relative_ndc_affine_mat'
        results['relative_ndc_affine_mat'] = theta2_ndc

        return results

    @cache_randomness
    def _apply_random_affine_to_theta(self, theta: np.ndarray,
                                      translates: Sequence[float],
                                      zoom: Sequence[float],
                                      shear: Sequence[float],
                                      rotate: Sequence[float]) -> np.ndarray:
        """Get the 3 x 3 affine transformation matrix in normal device
        coordinates based on input transformation matrix and transformation
        dict.

        Args:
            translates (list): the translation ratios along x axis and y axis.
                Defaults to(0., 0.).
            zoom (list): the min and max zoom ratios. Defaults to (1.0, 1.0).
            shear (list): the min and max shear ratios. Defaults to (1.0, 1.0).
            rotate (list): the min and max rotate degree. Defaults to (0., 0.).

        Returns:
            ndarray: affine transformation matrix.
        """

        valid = False

        while not valid:

            zoom_ = np.random.uniform(zoom[0], zoom[1])
            shear_ = np.random.uniform(shear[0], shear[1])

            t_x = np.random.uniform(-translates[0], translates[0])
            t_y = np.random.uniform(-translates[1], translates[1])

            phi = np.random.uniform(rotate[0] * np.pi / 180.,
                                    rotate[1] * np.pi / 180.)

            sin_phi = np.sin(phi)
            cos_phi = np.cos(phi)

            translate_mat = np.array([
                [1., 0., t_x],
                [0., 1., t_y],
                [0., 0., 1.],
            ])

            rotate_mat = np.array([
                [cos_phi, -sin_phi, 0.],
                [sin_phi, cos_phi, 0.],
                [0., 0., 1.],
            ])

            shear_mat = np.array([
                [shear_, 0., 0.],
                [0., 1. / shear_, 0.],
                [0., 0., 1.],
            ])

            zoom_mat = np.array([
                [zoom_, 0., 0.],
                [0., zoom_, 0.],
                [0., 0., 1.],
            ])

            T = translate_mat @ rotate_mat @ shear_mat @ zoom_mat

            theta_propose = T @ theta

            if not self.preserve_valid:
                break

            valid = theta_is_valid(theta_propose)

        return theta_propose

    def __repr__(self):
        return (f'{self.__class__.__name__}'
                f'(global_transform={self.global_transform}, '
                f'relative_transform={self.relative_transform}, '
                f'preserve_valid={self.preserve_valid}, '
                f'check_bound={self.check_bound})')



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional, Union

import torch
from torch import Tensor

from mmflow.registry import MODELS
from mmflow.utils import OptSampleList, SampleList, TensorDict
from ..builder import build_flow_estimator
from ..utils import BasicLink
from .base_flow_estimator import FlowEstimator


@MODELS.register_module()
class FlowNetCSS(FlowEstimator):
    """FlowNet2CSS model.

    Args:
        flownetC (dict): The config for FlownetC estimator.
        flownetS1 (dict): The first config for FlownetS estimator.
        flownetS2 (dict, optional): The second config for FlownetS estimator.
        link_cfg (dict): The config dict of link used to connect the
            estimators. Default to dict(scale_factor=4, mode='bilinear').
        flow_div (float): The divisor used to scale down ground truth.
            Defaults to 20.
        out_level (str): The level of output flow. Default to 'level2'.
        init_cfg (dict, list, optional): Config of dict weights initialization.
            Default: None.
    """

    def __init__(self,
                 flownetC: dict,
                 flownetS1: dict,
                 flownetS2: Optional[dict] = None,
                 link_cfg: dict = dict(scale_factor=4, mode='bilinear'),
                 flow_div: float = 20.,
                 out_level: str = 'level2',
                 init_cfg: Optional[Union[list, dict]] = None,
                 **kwargs) -> None:

        super().__init__(init_cfg=init_cfg, **kwargs)

        self.flownetC = build_flow_estimator(flownetC)
        self.flownetS1 = build_flow_estimator(flownetS1)
        self.link = BasicLink(**link_cfg)
        self.flow_div = flow_div

        if flownetS2 is not None:
            self.flownetS2 = build_flow_estimator(flownetS2)

        self.out_level = out_level

    def extract_feat(self, inputs: Tensor) -> TensorDict:
        """Extract features from images."""
        flowc = self.flownetC(inputs, mode='tensor')[self.out_level]

        img_channels = inputs.shape[1] // 2
        img1 = inputs[:, :img_channels, ...]
        img2 = inputs[:, img_channels:, ...]

        link_output1 = self.link(img1, img2, flowc, self.flow_div)

        concat1 = torch.cat(
            (img1, img2, link_output1.warped_img2, link_output1.upsample_flow,
             link_output1.brightness_err),
            dim=1)
        return concat1, img1, img2

    def loss(self, inputs: Tensor, data_samples: SampleList) -> dict:
        """Forward function for Flownet2CSS when model training.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The losses of output.
        """

        concat1, img1, img2 = self.extract_feat(inputs)

        # Train FlowNetCS before FlowNetCSS
        if hasattr(self, 'flownetS2'):
            flows1 = self.flownetS1(concat1, mode='tensor')[self.out_level]

            link_output2 = self.link(img1, img2, flows1, self.flow_div)
            concat2 = torch.cat((
                img1,
                img2,
                link_output2.warped_img2,
                link_output2.upsample_flow,
                link_output2.brightness_err,
            ),
                                dim=1)
            # when flownetS2 does not have flow_loss, loss if the multi-levels
            # flow_pred
            loss = self.flownetS2(concat2, data_samples, mode='loss')
        else:
            loss = self.flownetS1(concat1, data_samples, mode='loss')

        return loss

    def predict(self,
                inputs: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function for Flownet2CSS when model testing.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """
        concat1, img1, img2 = self.extract_feat(inputs)

        # Train FlowNetCS before FlowNetCSS
        if hasattr(self, 'flownetS2'):
            flows1 = self.flownetS1(concat1, mode='tensor')[self.out_level]

            link_output2 = self.link(img1, img2, flows1, self.flow_div)
            concat2 = torch.cat(
                (img1, img2, link_output2.warped_img2,
                 link_output2.upsample_flow, link_output2.brightness_err),
                dim=1)
            flow_result = self.flownetS2(concat2, data_samples, mode='predict')
        else:
            flow_result = self.flownetS1(concat1, data_samples, mode='predict')

        return flow_result

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> TensorDict:
        """Forward function of Flownet2CSS.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Dict[str, Tensor]: The predicted multi-level optical flow.
        """
        concat1, img1, img2 = self.extract_feat(inputs)

        # Train FlowNetCS before FlowNetCSS
        if hasattr(self, 'flownetS2'):
            flows1 = self.flownetS1(concat1, mode='tensor')[self.out_level]

            link_output2 = self.link(img1, img2, flows1, self.flow_div)
            concat2 = torch.cat(
                (img1, img2, link_output2.warped_img2,
                 link_output2.upsample_flow, link_output2.brightness_err),
                dim=1)
            flow_result = self.flownetS2(concat2, data_samples, mode='tensor')
        else:
            flow_result = self.flownetS1(concat1, data_samples, mode='tensor')
        return flow_result


@MODELS.register_module()
class FlowNet2(FlowEstimator):
    """FlowNet2 model.

    Args:
        flownetCSS (dict): The config of FlowNet2CSS estimator.
        flownetSD (dict): The config of FlowNet2SD estimator.
        flownet_fusion (dict): The config of fusion flownet.
        link_cfg (dict): The config dict of link used to connect the
            estimators. Defaults to dict(scale_factor=4, mode='nearest').
        flow_div (float): The divisor used to scale down ground truth.
            Defaults to 20.
        out_level (str): The level of output flow. Default to 'level2'.
        init_cfg (dict, list, optional): Config of dict weights initialization.
            Default: None.
    """

    def __init__(self,
                 flownetCSS: dict,
                 flownetSD: dict,
                 flownet_fusion: dict,
                 link_cfg: dict = dict(scale_factor=4, mode='nearest'),
                 flow_div: float = 20.,
                 out_level: str = 'level2',
                 init_cfg: Optional[Union[list, dict]] = None,
                 **kwargs) -> None:
        super().__init__(init_cfg=init_cfg, **kwargs)

        self.flownetCSS = build_flow_estimator(flownetCSS)
        self.flownetSD = build_flow_estimator(flownetSD)
        self.flownet_fusion = build_flow_estimator(flownet_fusion)
        self.link = BasicLink(**link_cfg)
        self.flow_div = flow_div

        self.out_level = out_level

    def loss(self, inputs: Tensor, data_samples: SampleList) -> dict:
        """Forward function for FlowNet2 when model training.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`]): Each item contains the
                meta information of each image and corresponding annotations.

        Returns:
            Dict[str, Tensor]: The losses of output.
        """
        loss = self.flownet_fusion(
            self.extract_feat(inputs), data_samples, mode='loss')

        return loss

    def predict(self,
                inputs: Tensor,
                data_samples: OptSampleList = None) -> SampleList:
        """Forward function for FlowNet2 when model testing.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Sequence[FlowDataSample]: The batch of predicted optical flow
                with the same size of images before augmentation.
        """

        return self.flownet_fusion(
            self.extract_feat(inputs), data_samples, mode='predict')

    def _forward(self,
                 inputs: Tensor,
                 data_samples: OptSampleList = None) -> TensorDict:
        """Forward function of FlowNet2.

        Args:
            inputs (Tensor): The concatenated input images.
            data_samples (list[:obj:`FlowDataSample`], optional): Each item
                contains the meta information of each image and corresponding
                annotations. Defaults to None.

        Returns:
            Dict[str, Tensor]: The predicted multi-level optical flow.
        """
        return self.flownet_fusion(
            self.extract_feat(inputs), data_samples, mode='tensor')

    def extract_feat(self, inputs: Tensor) -> TensorDict:
        img_channels = inputs.shape[1] // 2
        img1 = inputs[:, :img_channels, ...]
        img2 = inputs[:, img_channels:, ...]

        flow_css = self.flownetCSS._forward(inputs)[self.out_level]
        flow_sd = self.flownetSD._forward(inputs)[self.out_level]

        link_output_css = self.link(img1, img2, flow_css, self.flow_div)
        link_output_sd = self.link(img1, img2, flow_sd, self.flow_div)

        concat_feat = torch.cat(
            (img1, link_output_sd.scaled_flow, link_output_css.scaled_flow,
             link_output_sd.norm_scaled_flow, link_output_css.norm_scaled_flow,
             link_output_sd.brightness_err, link_output_css.brightness_err),
            dim=1)
        return concat_feat



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Sequence

import torch
import torch.nn as nn
from torch import Tensor

from mmflow.registry import MODELS


def sequence_loss(preds, flow_gt, gamma, valid=None, max_flow=400):
    """Compute sequence loss between prediction and ground truth.

    Args:
        preds (list(torch.Tensor)): List of flow prediction from
            flow_estimator.
        flow_gt (torch.Tensor): Ground truth flow map.
        gamma (float): Scale factor gamma in loss calculation.
        valid (torch.Tensor, optional): Tensor Used to exclude invalid pixels.
            Default: None.
        max_flow (int, optional): Used to exclude extremely large
            displacements. Default: 400.

    Returns:
        flow_loss (float): Total sequence loss.
    """
    n_preds = len(preds)
    flow_loss = 0.
    mag = torch.sum(flow_gt**2, dim=1).sqrt()
    if valid is None:
        valid = torch.ones(flow_gt[:, 0, :, :].shape).to(flow_gt)
    else:
        valid = ((valid >= 0.5) & (mag < max_flow)).to(flow_gt)

    for i, pred in enumerate(preds):
        i_weight = gamma**(n_preds - i - 1)
        i_loss = (pred - flow_gt).abs()
        flow_loss += i_weight * (valid[:, None] * i_loss).mean()
    return flow_loss


@MODELS.register_module()
class SequenceLoss(nn.Module):
    """Sequence Loss for RAFT.

    Args:
        gamma (float): The base of exponentially increasing weights. Default to
            0.8.
        max_flow (float): The maximum value of optical flow, if some pixel's
            flow of target is larger than it, this pixel is not valid.
                Default to 400.
    """

    def __init__(self, gamma: float = 0.8, max_flow: float = 400.) -> None:
        super().__init__()
        self.gamma = gamma
        self.max_flow = max_flow

    def forward(self,
                flow_preds: Sequence[Tensor],
                flow_gt: Tensor,
                valid: Tensor = None) -> Tensor:
        """Forward function for MultiLevelEPE.

        Args:
            preds_dict Sequence[Tensor]: The list of predicted optical flow.
            target (Tensor): Ground truth of optical flow with shape
                (B, 2, H, W).
            valid (Tensor, optional): Valid mask for optical flow.
                Defaults to None.

        Returns:
            Tensor: value of pixel-wise end point error loss.
        """
        return sequence_loss(flow_preds, flow_gt, self.gamma, valid,
                             self.max_flow)



_base_ = [
    '../_base_/models/flownet2/flownet2css.py',
    '../_base_/datasets/flyingchairs_384x448.py',
    '../_base_/schedules/schedule_s_long.py', '../_base_/default_runtime.py'
]
# Initialize weights of FlowNet2CS with
load_from = 'https://download.openmmlab.com/mmflow/flownet2/flownet2cs_8x1_sfine_flyingthings3d_subset_384x768.pth'  # noqa



# Copyright (c) OpenMMLab. All rights reserved.
from math import sqrt

import torch
from mmcv.cnn import build_activation_layer
from mmengine.model import BaseModule
from mmengine.runner import autocast

from mmflow.models import build_operators


class CorrBlock(BaseModule):
    """Basic Correlation Block.

    A block used to calculate correlation.

    Args:
        corr (dict): Config dict for build correlation operator.
        act_cfg (dict): Config dict for activation layer.
        scaled (bool): Whether to use scaled correlation by the number of
            elements involved to calculate correlation or not.
            Default: False.
        scale_mode (str): How to scale correlation. The value includes
        `'dimension'` and `'sqrt dimension'`, but it doesn't work when
        scaled = True. Default to `'dimension'`.
    """

    def __init__(self,
                 corr_cfg: dict,
                 act_cfg: dict = dict(type='LeakyReLU', negative_slope=0.1),
                 scaled: bool = False,
                 scale_mode: str = 'dimension') -> None:
        super().__init__()

        assert scale_mode in ('dimension', 'sqrt dimension'), (
            'scale_mode must be \'dimension\' or \'sqrt dimension\' '
            f'but got {scale_mode}')

        corr = build_operators(corr_cfg)
        act = build_activation_layer(act_cfg)
        self.scaled = scaled
        self.scale_mode = scale_mode

        self.kernel_size = corr.kernel_size
        self.corr_block = [corr, act]
        self.stride = corr_cfg.get('stride', 1)

    def forward(self, feat1: torch.Tensor,
                feat2: torch.Tensor) -> torch.Tensor:
        """Forward function for CorrBlock.

        Args:
            feat1 (Tensor): The feature from the first image.
            feat2 (Tensor): The feature from the second image.

        Returns:
            Tensor: The correlation between feature1 and feature2.
        """
        N, C, H, W = feat1.shape
        scale_factor = 1.

        if self.scaled:

            if 'sqrt' in self.scale_mode:
                scale_factor = sqrt(float(C * self.kernel_size**2))
            else:
                scale_factor = float(C * self.kernel_size**2)

        with autocast(enabled=False):
            corr = self.corr_block[0](feat1.float(),
                                      feat2.float()) / scale_factor

        corr = corr.view(N, -1, H // self.stride, W // self.stride)

        out = self.corr_block[1](corr)

        return out

    def __repr__(self):
        s = super().__repr__()
        s += f'\nscaled={self.scaled}'
        s += f'\nscale_mode={self.scale_mode}'
        return s



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.ops import Correlation
from torch.nn import Module

from mmflow.registry import MODELS

ENCODERS = MODELS
DECODERS = MODELS
FLOW_ESTIMATORS = MODELS
LOSSES = MODELS
COMPONENTS = MODELS
OPERATORS = MODELS

OPERATORS.register_module(module=Correlation)


def build_operators(cfg: dict) -> Module:
    """build opterator with config dict.

    Args:
        cfg (dict): The config dict of operator.

    Returns:
        Module: The built operator.
    """
    return OPERATORS.build(cfg)


def build_encoder(cfg: dict) -> Module:
    """Build encoder for flow estimator.

    Args:
        cfg (dict): Config for encoder.

    Returns:
        Module: Encoder module.
    """
    return ENCODERS.build(cfg)


def build_decoder(cfg: dict) -> Module:
    """Build decoder for flow estimator.

    Args:
        cfg (dict): Config for decoder.

    Returns:
        Module: Decoder module.
    """
    return DECODERS.build(cfg)


def build_components(cfg: dict) -> Module:
    """Build encoder for model component.

    Args:
        cfg (dict): Config for component of model.

    Returns:
        Module: Component of model.
    """
    return COMPONENTS.build(cfg)


def build_loss(cfg: dict) -> Module:
    """Build loss function.

    Args:
        cfg (dict): Config for loss function.

    Returns:
        Module: Loss function.
    """
    return LOSSES.build(cfg)


def build_flow_estimator(cfg: dict) -> Module:
    """Build flow estimator.

    Args:
        cfg (dict): Config for optical flow estimator.

    Returns:
        Module: Flow estimator.
    """
    return FLOW_ESTIMATORS.build(cfg)



# Copyright (c) OpenMMLab. All rights reserved.
import platform
from typing import Optional

from mmengine.config import Config
from torch.utils.data import Dataset

from mmflow.registry import DATASETS

if platform.system() != 'Windows':
    # https://github.com/pytorch/pytorch/issues/973
    import resource
    rlimit = resource.getrlimit(resource.RLIMIT_NOFILE)
    base_soft_limit = rlimit[0]
    hard_limit = rlimit[1]
    soft_limit = min(max(4096, base_soft_limit), hard_limit)
    resource.setrlimit(resource.RLIMIT_NOFILE, (soft_limit, hard_limit))


def build_dataset(cfg: Config, default_args: Optional[dict] = None) -> Dataset:
    """Build Pytorch dataset.

    Args:
        cfg (mmengine.Config): Config dict of dataset. It should at
            least contain the key "type".
        default_args (dict, optional): Default initialization arguments.

    Returns:
        dataset: The built dataset based on the input config.
    """
    dataset = DATASETS.build(cfg, default_args=default_args)

    return dataset



_base_ = [
    '../_base_/models/liteflownet2/liteflownet2_pre_M6S6.py',
    '../_base_/datasets/flyingchairs_320x448.py',
    '../_base_/schedules/liteflownet_pre_300k.py',
    '../_base_/default_runtime.py'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .flownet import FlowNetC, FlowNetS
from .flownet2 import FlowNet2, FlowNetCSS
from .irrpwc import IRRPWC
from .liteflownet import LiteFlowNet
from .maskflownet import MaskFlowNet, MaskFlowNetS
from .pwcnet import PWCNet
from .raft import RAFT

__all__ = [
    'FlowNetC', 'FlowNetS', 'LiteFlowNet', 'PWCNet', 'MaskFlowNetS', 'RAFT',
    'IRRPWC', 'FlowNet2', 'FlowNetCSS', 'MaskFlowNet'
]



dataset_type = 'FlyingChairs'
data_root = 'data/FlyingChairs_release'
# file_client_args = dict(
#     backend='petrel',
#     path_mapping=dict({
#         './data/': 's3://openmmlab/datasets/mmflow/',
#         'data/': 's3://openmmlab/datasets/mmflow/'
#     }))
file_client_args = dict(backend='disk')
train_pipeline = [
    dict(type='LoadImageFromFile', file_client_args=file_client_args),
    dict(type='LoadAnnotations', file_client_args=file_client_args),
    dict(
        type='ColorJitter',
        asymmetric_prob=0.2,
        brightness=0.4,
        contrast=0.4,
        saturation=0.4,
        hue=0.5 / 3.14),
    dict(type='Erase', prob=0.5, bounds=[50, 100], max_num=3),
    dict(
        type='SpacialTransform',
        spacial_prob=0.8,
        stretch_prob=0.8,
        crop_size=(368, 496),
        min_scale=-0.1,
        max_scale=1.,
        max_stretch=0.2),
    dict(type='RandomCrop', crop_size=(368, 496)),
    dict(type='RandomFlip', prob=0.5, direction='horizontal'),
    dict(type='RandomFlip', prob=0.1, direction='vertical'),
    dict(type='Validation', max_flow=1000.),
    dict(type='PackFlowInputs')
]
test_pipeline = [
    dict(type='LoadImageFromFile'),
    dict(type='LoadAnnotations'),
    dict(type='InputPad', exponent=3),
    dict(type='PackFlowInputs')
]

flyingchairs_train = dict(
    type=dataset_type,
    pipeline=train_pipeline,
    data_root=data_root,
    split_file='data/FlyingChairs_release/FlyingChairs_train_val.txt')
flyingchairs_test = dict(
    type=dataset_type,
    pipeline=test_pipeline,
    data_root=data_root,
    test_mode=True,
    split_file='data/FlyingChairs_release/FlyingChairs_train_val.txt')
train_dataloader = dict(
    batch_size=2,
    num_workers=4,
    sampler=dict(type='InfiniteSampler', shuffle=True),
    drop_last=True,
    persistent_workers=True,
    dataset=flyingchairs_train)
val_dataloader = dict(
    batch_size=1,
    num_workers=2,
    sampler=dict(type='DefaultSampler', shuffle=False),
    drop_last=False,
    persistent_workers=True,
    dataset=flyingchairs_test)
test_dataloader = val_dataloader
val_evaluator = dict(type='EndPointError')
test_evaluator = val_evaluator
